# coding=utf-8
from __future__ import unicode_literals, print_function, absolute_import

import collections
import datetime
import struct
import sys
import sysconfig

import pandas as pd
from typing import NoReturn, Text, Dict, Any, List, Set, Sequence, Union, Deque

from bw import __version__

from bw.constant import DATA_TYPE_TICK,BarLikeDict2, ObjectLikeDict

from bw.csdk.cpp_sdk import CTick
from bw.csdk.cpp_sdk import CQuote
from bw.csdk.cpp_sdk import COrder
from bw.csdk.cpp_sdk import CExecRpt
from bw.csdk.cpp_sdk import CCash
from bw.csdk.cpp_sdk import CPosition

from bw.csdk.cpp_sdk import CAccount
from bw.csdk.cpp_sdk import CAccountStatus
from bw.csdk.cpp_sdk import CParameter
from bw.csdk.cpp_sdk import CIndicator
from bw.csdk.cpp_sdk import CBar


from bw.csdk.cpp_sdk import py_bwi_get_cash, py_bwi_get_positions, \
    py_bwi_get_accounts, py_bwi_set_timer, py_bwi_now_plus
from bw.enum import MODE_BACKTEST, MODE_LIVE
from bw.yy.account import Account,Cash

from bw.model.account import YYAccount

from bw.utils import bwsdklogger, beijing_tzinfo,DictLikeObject,ObjectLikeDict


class BarSubInfo(object):
    __slots__ = ['symbol', 'frequency']

    def __init__(self, symbol, frequency):
        self.symbol = symbol  # type: Text
        self.frequency = frequency  # type: Text

    def __hash__(self):
        return hash((self.symbol, self.frequency))

    def __eq__(self, other):
        if not isinstance(other, BarSubInfo):
            return False

        return (self.symbol, self.frequency) == (other.symbol, other.frequency)


class BarWaitgroupInfo(object):
    __slots__ = ['symbols', 'frequency', 'timeout_seconds']

    def __init__(self, frequency, timeout_seconds):
        self.symbols = set()  # type: Set[Text]
        self.frequency = frequency  # type: Text
        self.timeout_seconds = timeout_seconds  # type: int

    def add_one_symbol(self, symbol):
        # type: (Text) ->NoReturn
        self.symbols.add(symbol)

    def add_symbols(self, syms):
        # type: (Sequence[Text]) ->NoReturn
        self.symbols.union(syms)

    def remove_one_symbol(self, symbol):
        # type: (Text) ->NoReturn
        self.symbols.discard(symbol)

    def is_symbol_in(self, symbol):
        # type: (Text) -> bool
        return symbol in self.symbols

    def __len__(self):
        return len(self.symbols)


class DefaultFileModule(object):
    def on_tick(self, ctx, tick):
        print('请初始化on_tick方法')

    def on_bar(self, ctx, bar):
        print('请初始化on_bar方法')

    def on_bod(self, ctx, bod):
        print('请初始化on_bod方法')

    def on_eod(self, ctx, eod):
        print('请初始化on_eod方法')


class Context(object):
    """
    策略运行的上下文类. 这在整个进程中要保证是单一实例
    注意: 一个运行的python进程只能运行一个策略.
    警告: 客户写的策略代码不要直接使用这个类来实例化, 而是使用 sdk 实例化好的 context 实例
    """
    inside_file_module = DefaultFileModule()
    on_bar_fun = None
    on_tick_fun = None
    init_fun = None
    on_execution_report_fun = None
    on_order_status_fun = None
    on_backtest_finished_fun = None
    on_parameter_fun = None
    on_error_fun = None
    on_shutdown_fun = None
    on_trade_data_connected_fun = None
    on_market_data_connected_fun = None
    on_account_status_fun = None
    on_market_data_disconnected_fun = None
    on_trade_data_disconnected_fun = None

    strategy_id = ''  # type: Text
    token = None  # type: Text
    mode = None  # type: int
    backtest_and_not_wait_group = False  # type: bool
    _temporary_now = None  # type: datetime.datetime  # 用于回测模式下, 且wait_group=True时, 要修改context.now 时的值, 因为这时时钟是走到了下一个时刻, 而数据还是上一时刻的, 所以要修改时钟为上一时刻,跟数据时间一致
    backtest_start_time = None  # type: Text
    backtest_end_time = None  # type: Text
    adjust_mode = None  # type: int
    inside_schedules = {}  # type: Dict[Text, Any]

    sdk_lang = "python{}.{}".format(sys.version_info.major, sys.version_info.minor)  # type: Text
    sdk_version = __version__.__version__  # type: Text
    sdk_arch = str(struct.calcsize("P") * 8)  # type: Text
    sdk_os = sysconfig.get_platform()  # type: Text

    tick_data_cache = dict()  # type: Dict[Text, Deque[Any]]
    max_tick_data_count = 1
    bar_data_cache = dict()  # type: Dict[Text, Deque[Any]]  # 以 bar.symbol+bar.frequency作为key
    max_bar_data_count = 1
    bar_data_set = set()  # type: Set  # 保存已有的bar的 (symbol, frequency, eob), 用于判断是否重复值
    tick_sub_symbols = set()  # type: Set[Text]   # 订阅tick的symbol
    bar_sub_infos = set()  # type: Set[BarSubInfo]   # 订阅bar的信息集合
    # 订阅bar用freequency做为key, 相应的股票集合做为value
    bar_waitgroup_frequency2Info = dict()  # type: Dict[Text, BarWaitgroupInfo]

    is_set_onbar_timeout_check = False

    def __init__(self):
        self.inside_accounts = {}  # type: Dict[Text, YYAccount]

    def _set_onbar_waitgroup_timeout_check(self):
        if self.is_live_model() and not self.is_set_onbar_timeout_check:
            # 实时模式下 3000毫秒触发一次timer事件 用来处理 wait_group 的过期.
            py_bwi_set_timer(3000)
            self.is_set_onbar_timeout_check = True

    def _add_bar2bar_data_cache(self, k, bar):
        # type: (Text, Dict[Text, Any]) -> NoReturn
        kk = (bar['symbol'], bar['frequency'], bar['eob'])
        if kk in self.bar_data_set:
            bwsdklogger.debug("bar data %s 已存在, 跳过不加入", kk)
        else:
            if k not in context.bar_data_cache:
                bwsdklogger.debug("bar data %s 在context.bar_data_cache相应的key=%s不存在, 加个key, deque长度为:%d", kk, k,
                                  context.max_bar_data_count)
                context.bar_data_cache[k] = collections.deque(maxlen=context.max_bar_data_count)

            context.bar_data_cache[k].appendleft(bar)
            self.bar_data_set.add(kk)

    @property
    def has_wait_group(self):
        # type: () ->bool
        return len(self.bar_waitgroup_frequency2Info) > 0

    @property
    def now(self):
        # type: ()->datetime.datetime
        """
        实时模式返回当前本地时间, 回测模式返回当前回测时间
        """
        if self._temporary_now:  # 这个是在回测模式且wait_group=True的情况下时存在的
            if self._temporary_now.tzinfo is None:
                return self._temporary_now.replace(tzinfo=beijing_tzinfo)
            return self._temporary_now

        now = py_bwi_now_plus()
        # now == 0 说明是回测模式而且处于init装填 c sdk拿不到时间
        if now == 0:
            return datetime.datetime.strptime(context.backtest_start_time, "%Y-%m-%d %H:%M:%S").replace(
                tzinfo=beijing_tzinfo)

        return datetime.datetime.fromtimestamp(now).replace(tzinfo=beijing_tzinfo)

    @property
    def symbols(self):
        # type: ()->Set[Text]
        """
        订阅bar跟tick的symbol集合
        bar 的symbols + tick 的symbols
        """
        return set(barsub.symbol for barsub in self.bar_sub_infos).union(self.tick_sub_symbols)

    @property
    def accounts(self):
        # type: ()->Dict[Text, Account]
        """
        用户资金 & 持仓情况
        """
        if not self.inside_accounts or len(self.inside_accounts)==0:
            # print('not in accounts')
            self._set_accounts()
        
        return self.inside_accounts

    def account(self, account_id=''):
        accounts = self.accounts
        # 只有一个账户 且未有account_id 的情况下返回唯一用户
        if not account_id and len(accounts) == 1:
            default_id = sorted(accounts.keys())[0]
            return accounts.get(default_id)

        return accounts.get(account_id)

    # @property
    # def parameters(self):
    #     """
    #     动态参数
    #     """
    #     from bw.api.basic import get_parameters
    #     parameters = get_parameters()
    #     return {p['key']: p for p in parameters}

    def _set_accounts(self):
        # print('first get accounts then do somthing')
        acnt= py_bwi_get_accounts()
        accountsls=[]
        for item in acnt:
            it = Account(item)
            accountsls.append(it)
            pass
        
        # print('%'*30,len(accountsls))

        # print(accountsls)
        if accountsls is None:
            return
        for account in accountsls:  # type: yyAccount
            # print('-*'*20,account.state)
            # print(account.account_id,account.account_name)
            self._get_account_info(account.account_id, account.account_name)

    @staticmethod
    def data(symbol, frequency, count=1, fields='', bar_df=True):
        # type: (Text, Text, int, Text, bool) -> Union[List[Union[ObjectLikeDict, BarLikeDict2]], pd.DataFrame]
        """
        获取订阅的bar或tick滑窗，数据为包含当前时刻推送bar或tick的前count条bar(tick)数据.
        注意: 为了兼容之前的sdk版本, 在 frequency 为非 tick 的情况下, 返回的数据类型为 pandas的DataFrame.
        你可以显示的设置参数 bar_df=False, 这样返回的的数据是 List[BarLikeDict2].
        frequency为tick时, 返回的数据类型为 List[ObjectLikeDict]
        fields 参数的取值情况:
          1. 当 frequency == 'tick' 时, fields的取值可以为:'quotes,symbol,created_at,price,open,high,low,cum_volume,cum_amount,cum_position,last_amount,last_volume,trade_type'
          2. 当 数据为bar时, fields的取值可以为: 'symbol,eob,bob,open,close,high,low,volume,amount,pre_close,position,frequency'
        """
        symbol = symbol.strip()
        if ',' in symbol:
            bwsdklogger.warning('不支持返回多个symbol, 返回空list')
            return []
        if ',' in frequency:
            bwsdklogger.warning('不支持多个频率, 返回空list')
            return []

        if count < 1:
            count = 1
        # frequency = adjust_frequency(frequency)
        if frequency == DATA_TYPE_TICK:
            # print('---------------------------------frequency == DATA_TYPE_TICK')
            if symbol in context.tick_data_cache:
                symbol_tick_queue = context.tick_data_cache[symbol]  # type: collections.deque
                if count > symbol_tick_queue.maxlen:
                    count = symbol_tick_queue.maxlen
                if count > len(symbol_tick_queue):
                    # 数据不够, 说明要补数据
                    delta_count = symbol_tick_queue.maxlen - len(symbol_tick_queue)
                    from bw.api.query import history_n
                    if len(symbol_tick_queue) == 0:
                        query_data_end_time = context.now
                    else:
                        query_data_end_time = symbol_tick_queue[-1]['created_at']
                    tick_datas = history_n(symbol=symbol, frequency=frequency, count=delta_count,
                                           end_time=query_data_end_time)
                    tick_datas_trans = list()  # type: List[ObjectLikeDict]
                    for bd in tick_datas:
                        tick_datas_trans.append(ObjectLikeDict(data=bd))

                    comb_tick_datas = list(symbol_tick_queue) + tick_datas_trans  # type: List[ObjectLikeDict]
                    # 排除掉重复的数据
                    tick_key_set = set()
                    new_tick_datas = list()  # type: List[ObjectLikeDict]
                    for tick_item in comb_tick_datas:
                        kk = (tick_item['symbol'], tick_item['created_at'])
                        if kk not in tick_key_set:
                            tick_key_set.add(kk)
                            new_tick_datas.append(tick_item)

                    new_tick_datas.sort(key=lambda item: item['created_at'])
                    symbol_tick_queue.clear()
                    symbol_tick_queue.extendleft(new_tick_datas)

                # 按要求取数据
                d = [v for idx, v in enumerate(symbol_tick_queue) if idx < count]
                d.reverse()
                if len(fields) == 0:
                    return d
                else:  # 用户指定字段的情况, 进行字段的筛选
                    all_fields = {'quotes', 'symbol', 'created_at', 'price', 'open', 'high', 'low', 'cum_volume',
                                  'cum_amount', 'cum_position', 'last_amount', 'last_volume', 'trade_type'}
                    filter_fields = []
                    for f in fields.lower().strip().split(','):
                        f = f.strip()
                        if f in all_fields:
                            filter_fields.append(f)

                    if len(filter_fields) == 0:
                        return []

                    dd = []
                    for bar_data_item in d:
                        item = {}
                        for f in filter_fields:
                            item[f] = bar_data_item.data.get(f)
                        dd.append(ObjectLikeDict(data=item))
                    return dd
            else:
                return []
        else:  # bar type
            
            # print('---------------------------------frequency !!!!=!!! DATA_TYPE_TICK')
            all_fields = {'symbol', 'eob', 'bob', 'open', 'close', 'high', 'low', 'volume', 'amount', 'pre_close',
                          'position', 'frequency', }
            filter_fields = []
            for f in fields.lower().strip().split(','):
                f = f.strip()
                if f in all_fields:
                    filter_fields.append(f)

            k = symbol + "_" + frequency
            if k in context.bar_data_cache:
                symbol_bar_queue = context.bar_data_cache[k]  # type: collections.deque
                if count > symbol_bar_queue.maxlen:
                    count = symbol_bar_queue.maxlen
                if count > len(symbol_bar_queue):
                    # 数据不够, 说明要补数据
                    delta_count = symbol_bar_queue.maxlen - len(symbol_bar_queue) + 1
                    from bw.api.query import history_n
                    if len(symbol_bar_queue) == 0:
                        query_data_end_time = context.now
                    else:
                        query_data_end_time = symbol_bar_queue[-1]['eob']
                    bar_datas = history_n(symbol=symbol, frequency=frequency, count=delta_count,
                                          end_time=query_data_end_time)

                    bar_datas_trans = list()  # type: List[BarLikeDict2]
                    for bd in bar_datas:
                        if frequency == '1d':  # 1d的话统一替换成 15:15:01
                            bd['eob'] = bd['eob'].replace(hour=15, minute=15, second=1)
                        bar_datas_trans.append(ObjectLikeDict (bd))

                    comb_bar_datas = list(symbol_bar_queue) + bar_datas_trans  # type: List[BarLikeDict2]
                    # 排除掉重复的数据
                    bar_key_set = set()
                    new_bar_datas = list()  # type: List[BarLikeDict2]
                    for bar_item in comb_bar_datas:
                        kk = (bar_item['symbol'], bar_item['frequency'], bar_item['eob'])
                        if kk not in bar_key_set:
                            bar_key_set.add(kk)
                            new_bar_datas.append(bar_item)
                            context.bar_data_set.add(ObjectLikeDict(bar_item))

                    new_bar_datas.sort(key=lambda item: item['eob'])
                    symbol_bar_queue.clear()
                    symbol_bar_queue.extendleft(new_bar_datas)

                # 按要求取数据
                d = [v for idx, v in enumerate(symbol_bar_queue) if idx < count]  # type: List[BarLikeDict2]
                d.reverse()
                if bar_df:
                    if len(fields) == 0:
                        td = [{
                            'symbol': item.symbol,
                            'eob': item.eob,
                            'bob': item.bob,
                            'open': item.open,
                            'close': item.close,
                            'high': item.high,
                            'low': item.low,
                            'amount': item.amount,
                            'pre_close': item.pre_close,
                            'position': item.position,
                            'frequency': item.frequency,
                            'volume': item.volume,
                        } for item in d]
                        return pd.DataFrame(data=td)
                    else:  # 用户指定字段的情况
                        dd = []
                        for tick_item in d:
                            item = {}
                            for f in filter_fields:
                                item[f] = tick_item.data.get(f)
                            dd.append(item)
                        return pd.DataFrame(data=dd)
                else:
                    if len(fields) == 0:
                        return d
                    else:  # 用户指定字段的情况
                        dd = []
                        for tick_item in d:
                            item = {}
                            for f in filter_fields:
                                item[f] = tick_item.data.get(f)
                            dd.append(ObjectLikeDict(item))
                        return dd
            else:
                if bar_df:
                    if len(fields) == 0:
                        return pd.DataFrame(
                            columns=['symbol', 'eob', 'bob', 'open', 'close', 'high', 'low', 'volume', 'amount',
                                     'pre_close', 'position', 'frequency'])
                    else:  # 用户指定字段的情况
                        return pd.DataFrame(columns=filter_fields)
                else:
                    return []

    def _get_account_info(self, account_id, account_name):
        # 资金信息
        ccash = py_bwi_get_cash(account_id)

        # print("*&"*30,len(ccash))

        if ccash is None:
            # print('cash is none')
            return
        if len(ccash)>0:
            cash = Cash(ccash[0])

        # 持仓信息
        positions = py_bwi_get_positions(account_id)
        if positions is None:
            # print('pos is none')
            return

        positions_infos = {'{}.{}'.format(position['symbol'], position['side']): position for position in positions}

        account = YYAccount(account_id, account_name, cash, positions_infos)
        # print('_get_account_info:',account)
        self.inside_accounts[account_id] = account

    def is_backtest_model(self):
        # type: () ->bool
        """
        是否回测模式
        """
        return self.mode == MODE_BACKTEST

    def is_live_model(self):
        # type: () ->bool
        """
        是否实时模式
        """
        return self.mode == MODE_LIVE


# 提供给API的唯一上下文实例
context = Context()



# coding=utf-8
from __future__ import unicode_literals, print_function, absolute_import

import asyncio
__title__ = 'bw'
__description__ = '牛富量化 python3 x64 sdk'
__url__ = 'http://bwsoft.cn/'
__version__ = '1.5.86'
__author__ = 'fengcl'
__author_email__ = '439126089@qq.com'
__license__ = 'Apache 2.0'



# coding=utf-8
from __future__ import unicode_literals

from typing import Text, Dict


# nipasbc
class Property(object):
    key = ...  # type: Text
    val = ...  # type: Text
    name = ...  # type: Text
    index = ...  # type: int
    visible = ...  # type: bool

    def __init__(self,
                 key: Text = None,
                 val: Text = None,
                 name: Text = None,
                 index: int = None,
                 visible: bool = None
                 ): ...


# nipasbc
class Filter(object):
    fields = ...  # type: Text
    filter = ...  # type: Text
    sort = ...  # type: Text
    limit = ...  # type: int
    page = ...  # type: int
    pagesize = ...  # type: int
    fromdate = ...  # type: Text
    todate = ...  # type: Text

    def __init__(self, fields: Text = None, filter: Text = None, sort: Text = None, limit: int = None, page: int = None,
                 pagesize: int = None, fromdate: Text = None, todate: Text = None): ...


# nipasbc
class Error(object):
    code = ...  # type: int
    type = ...  # type: Text
    info = ...  # type: Text

    def __init__(self, code: int = None, type: Text = None, info: Text = None): ...


# nipasbc
class ConnectionAddress(object):
    title = ...  # type: Text
    address = ...  # type: Dict[Text, Text]

    def __init__(self, title: Text = None, address: Dict[Text, Text] = None): ...


# nipasbc
class ConnectionStatus(object):
    State_UNKNOWN = 0  # 未知
    State_CONNECTING = 1  # 连接中
    State_CONNECTED = 2  # 已连接
    State_LOGGEDIN = 3  # 已登录
    State_DISCONNECTING = 4  # 断开中
    State_DISCONNECTED = 5  # 已断开
    State_ERROR = 6  # 错误

    state = ...  # type: int
    error = ...  # type: Error

    def __init__(self, state: int = None, error: Error = None): ...


# nipasbc
class Log(object):
    source = ...  # type: Text
    level = ...  # type: Text
    msg = ...  # type: Text
    owner_id = ...  # type: Text
    created_at = ...  # type: int

    # def __init__(self, source: Text = None, level: Text = None, msg: Text = None, owner_id: Text = None,
    #              created_at: int = None): ...


# nipasbc
class Logs(object):
    data = ...  # type: List[Log]

    # def __init__(self, data: List[Log] = None): ...


# nipasbc
class Heartbeat(object):
    created_at = ...  # type: int

    # def __init__(self, created_at: int = None): ...



class EnumTypeWrapper(object):
  """A utility for finding the names of enum values."""

  DESCRIPTOR = None

  def __init__(self, enum_type):
    """Inits EnumTypeWrapper with an EnumDescriptor."""
    self._enum_type = enum_type
    self.DESCRIPTOR = enum_type

  def Name(self, number):
    """Returns a string containing the name of an enum value."""
    if number in self._enum_type.values_by_number:
      return self._enum_type.values_by_number[number].name
    raise ValueError('Enum %s has no name defined for value %d' % (
        self._enum_type.name, number))

  def Value(self, name):
    """Returns the value corresponding to the given enum name."""
    if name in self._enum_type.values_by_name:
      return self._enum_type.values_by_name[name].number
    raise ValueError('Enum %s has no value defined for name %s' % (
        self._enum_type.name, name))

  def keys(self):
    """Return a list of the string names in the enum.
    These are returned in the order they were defined in the .proto file.
    """

    return [value_descriptor.name
            for value_descriptor in self._enum_type.values]

  def values(self):
    """Return a list of the integer values in the enum.
    These are returned in the order they were defined in the .proto file.
    """

    return [value_descriptor.number
            for value_descriptor in self._enum_type.values]

  def items(self):
    """Return a list of the (name, value) pairs of the enum.
    These are returned in the order they were defined in the .proto file.
    """
    return [(value_descriptor.name, value_descriptor.number)
            for value_descriptor in self._enum_type.values]

  def __getattr__(self, name):
    """Returns the value corresponding to the given enum name."""
    if name in self._enum_type.values_by_name:
      return self._enum_type.values_by_name[name].number
    raise AttributeError


from typing import Text, Dict

from bw.yy.account import Position



# nipasbc
class Indicator(object):
    # 账号ID
    # account_id = ...  # type: Text
    # # 累计收益率(pnl/cum_inout)
    # pnl_ratio = ...  # type: float
    # # 年化收益率
    # pnl_ratio_annual = ...  # type: float
    # # 夏普比率
    # sharp_ratio = ...  # type: float
    # # 最大回撤
    # max_drawdown = ...  # type: float
    # # 风险比率
    # risk_ratio = ...  # type: float
    # # 开仓次数
    # open_count = ...  # type: int
    # # 平仓次数
    # close_count = ...  # type: int
    # # 盈利次数
    # win_count = ...  # type: int
    # # 亏损次数
    # lose_count = ...  # type: int
    # # 胜率
    # win_ratio = ...  # type: float
    # # 卡玛比率
    # calmar_ratio = ...  # type: float
    # created_at = ...  # type: int
    # updated_at = ...  # type: int

    def __init__(self):
        self.account_id = ...  # type: Text
        # 累计收益率(pnl/cum_inout)
        self.pnl_ratio = ...  # type: float
        # 年化收益率
        self.pnl_ratio_annual = ...  # type: float
        # 夏普比率
        self.sharp_ratio = ...  # type: float
        # 最大回撤
        self.max_drawdown = ...  # type: float
        # 风险比率
        self.risk_ratio = ...  # type: float
        # 开仓次数
        self.open_count = ...  # type: int
        # 平仓次数
        self.close_count = ...  # type: int
        # 盈利次数
        self.win_count = ...  # type: int
        # 亏损次数
        self.lose_count = ...  # type: int
        # 胜率
        self.win_ratio = ...  # type: float
        # 卡玛比率
        self.calmar_ratio = ...  # type: float
        self.created_at = ...  # type: int
        self.updated_at = ...  # type: int
        
    def __init__(self,
                 account_id: Text = '',
                 pnl_ratio: float = None,
                 pnl_ratio_annual: float = None,
                 sharp_ratio: float = None,
                 max_drawdown: float = None,
                 risk_ratio: float = None,
                 open_count: int = None,
                 close_count: int = None,
                 win_count: int = None,
                 lose_count: int = None,
                 win_ratio: float = None,
                 calmar_ratio: float = None,
                 alpha: float  = None,
                 beta: float  = None,
                 sortino_rate: float  = None,
                 information_ratio: float  = None,
                 created_at: int = None,
                 updated_at: int = None,
                 bench_ratio : float  = None ,
                 bench_annul_ratio  : float  = None,
                 win_amount   : float  = None,   
                 loss_amount      : float  = None,  
                 win_loss_ratio    : float  = None, 

                 ): ...
    # def __init__(self,ind_):
    #     self.account_id=ind_.account_id#             //账号ID
    #     self.pnl_ratio=ind_.pnl_ratio#                      //累计收益率(const pnl/cum_inout)
    #     self.pnl_ratio_annual=ind_.pnl_ratio_annual#               //年化收益率
    #     self.sharp_ratio=ind_.sharp_ratio#                    //夏普比率
    #     self.max_drawdown=ind_.max_drawdown#                   //最大回撤
    #     self.risk_ratio=ind_.risk_ratio#                     //风险比率
    #     self.open_count=ind_.open_count#                     //开仓次数
    #     self.close_count=ind_.close_count#                    //平仓次数
    #     self.win_count=ind_.win_count#                      //盈利次数
    #     self.lose_count=ind_.lose_count#                     //亏损次数
    #     self.win_ratio=ind_.win_ratio#                      //胜率
    #     self.created_at=ind_.created_at#                    //指标创建时间
    #     self.updated_at=ind_.updated_at#                    //指标变更时间
    #     pass
    def FromFmt(self,ind_):
        self.account_id=ind_.account_id#             //账号ID
        self.pnl_ratio=ind_.pnl_ratio#                      //累计收益率(const pnl/cum_inout)
        self.pnl_ratio_annual=ind_.pnl_ratio_annual#               //年化收益率
        self.sharp_ratio=ind_.sharp_ratio#                    //夏普比率
        self.max_drawdown=ind_.max_drawdown#                   //最大回撤
        self.risk_ratio=ind_.risk_ratio#                     //风险比率
        self.open_count=ind_.open_count#                     //开仓次数
        self.close_count=ind_.close_count#                    //平仓次数
        self.win_count=ind_.win_count#                      //盈利次数
        self.lose_count=ind_.lose_count#                     //亏损次数
        self.win_ratio=ind_.win_ratio#                      //胜率
        self.created_at=ind_.created_at#                    //指标创建时间
        self.updated_at=ind_.updated_at#                    //指标变更时间
        
        self.alpha=ind_.alpha#                      //胜率
        self.beta=ind_.beta#                      //胜率
        self.sortino_rate=ind_.sortino_rate#                      //胜率
        self.information_ratio=ind_.information_ratio#                      //胜率

        self.bench_ratio= ind_.bench_ratio
        self.bench_annul_ratio=ind_.bench_annul_ratio
        self.win_amount=ind_.win_amount
        self.loss_amount=ind_.loss_amount
        self.win_loss_ratio=ind_.win_loss_ratio

        pass  

    def __str__(self):
        ls=[]
        ls.append("账号ID:       " + str(self.account_id )         )
        ls.append("累计收益率:   " + str(self.pnl_ratio )          )
        ls.append("年化收益率:   " + str(self.pnl_ratio_annual )   )
        ls.append("夏普比率:     " + str(self.sharp_ratio )        )
        ls.append("最大回撤:     " + str(self.max_drawdown )       )
        ls.append("风险比率:     " + str(self.risk_ratio )         )
        ls.append("开仓次数:     " + str(self.open_count )         )
        ls.append("平仓次数:     " + str(self.close_count )        )
        ls.append("盈利次数:     " + str(self.win_count )          )
        ls.append("亏损次数:     " + str(self.lose_count )         )

        ls.append("阿尔法:         " + str(self.alpha )          )
        ls.append("贝塔:         " + str(self.beta )          )
        ls.append("索提诺比:         " + str(self.sortino_rate )          )
        ls.append("信息比率:         " + str(self.information_ratio )          )

        ls.append("胜率:         " + str(self.win_ratio )          )
        ls.append("指标创建时间: " + str(self.created_at )         )
        ls.append("指标变更时间: " + str(self.updated_at )         )

        
        ls.append("基准: " +str(self.bench_ratio ))
        ls.append("基准年化: " +str(self.bench_annul_ratio ))
        ls.append("盈利数额: " +str(self.win_amount ))
        ls.append("亏损数额: " +str(self.loss_amount ))
        ls.append("盈亏比: " +str(self.win_loss_ratio ))
        
        return '\n'.join(ls)
        pass


# nipasbc
class Indicators(object):
    data = ...  # type: List[Indicator]

    # def __init__(self, data: List[Indicator] = None): ...


# nipasbc
class IndicatorDuration(object):
    # 账号ID
    account_id = ...  # type: Text
    pnl_ratio = ...  # type: float
    pnl = ...  # type: float
    fpnl = ...  # type: float
    frozen = ...  # type: float
    cash = ...  # type: float
    nav = ...  # type: float
    positions = ...  # type: List[Position]
    # 周期累计盈亏
    cum_pnl = ...  # type: float
    # 周期累计买入额
    cum_buy = ...  # type: float
    # 周期累计卖出额
    cum_sell = ...  # type: float
    # 周期累计手续费
    cum_commission = ...  # type: float
    duration = ...  # type: Duration
    created_at = ...  # type: int
    updated_at = ...  # type: int



    # def __init__(self,
    #              account_id: Text = None,
    #              pnl_ratio: float = None,
    #              pnl: float = None,
    #              fpnl: float = None,
    #              frozen: float = None,
    #              cash: float = None,
    #              nav: float = None,
    #              positions: list[Position] = None,
    #              cum_pnl: float = None,
    #              cum_buy: float = None,
    #              cum_sell: float = None,
    #              cum_commission: float = None,
    #              duration: Duration = None,
    #              created_at: int = None,
    #              updated_at: int = None
    #              ): ...


# nipasbc
class IndicatorDurations(object):
    data = ...  # type: List[IndicatorDuration]

    # def __init__(self, data: List[IndicatorDuration] = None): ...



# coding=utf8
from __future__ import print_function, absolute_import, unicode_literals

ExecType_Unknown = 0
ExecType_New = 1  # 已报
ExecType_DoneForDay = 4  #
ExecType_Canceled = 5  # 已撤销
ExecType_PendingCancel = 6  # 待撤销
ExecType_Stopped = 7  #
ExecType_Rejected = 8  # 已拒绝
ExecType_Suspended = 9  # 挂起
ExecType_PendingNew = 10  # 待报
ExecType_Calculated = 11  #
ExecType_Expired = 12  # 过期
ExecType_Restated = 13  #
ExecType_PendingReplace = 14  #
ExecType_Trade = 15  # 成交
ExecType_TradeCorrect = 16  #
ExecType_TradeCancel = 17  #
ExecType_OrderStatus = 18  # 委托状态
ExecType_CancelRejected = 19  # 撤单被拒绝

OrderStatus_Unknown = 0
OrderStatus_New = 1  # 已报
OrderStatus_PartiallyFilled = 2  # 部成
OrderStatus_Filled = 3  # 已成
OrderStatus_DoneForDay = 4  #
OrderStatus_Canceled = 5  # 已撤
OrderStatus_PendingCancel = 6  # 待撤
OrderStatus_Stopped = 7  #
OrderStatus_Rejected = 8  # 已拒绝
OrderStatus_Suspended = 9  # 挂起
OrderStatus_PendingNew = 10  # 待报
OrderStatus_Calculated = 11  #
OrderStatus_Expired = 12  # 已过期
OrderStatus_AcceptedForBidding = 13  #
OrderStatus_PendingReplace = 14  #

OrderRejectReason_Unknown = 0  # 未知原因
OrderRejectReason_RiskRuleCheckFailed = 1  # 不符合风控规则
OrderRejectReason_NoEnoughCash = 2  # 资金不足
OrderRejectReason_NoEnoughPosition = 3  # 仓位不足
OrderRejectReason_IllegalAccountId = 4  # 非法账户ID
OrderRejectReason_IllegalStrategyId = 5  # 非法策略ID
OrderRejectReason_IllegalSymbol = 6  # 非法交易标的
OrderRejectReason_IllegalVolume = 7  # 非法委托量
OrderRejectReason_IllegalPrice = 8  # 非法委托价
OrderRejectReason_AccountDisabled = 10  # 交易账号被禁止交易
OrderRejectReason_AccountDisconnected = 11  # 交易账号未连接
OrderRejectReason_AccountLoggedout = 12  # 交易账号未登录
OrderRejectReason_NotInTradingSession = 13  # 非交易时段
OrderRejectReason_OrderTypeNotSupported = 14  # 委托类型不支持
OrderRejectReason_Throttle = 15  # 流控限制

CancelOrderRejectReason_OrderFinalized = 101  # 委托已完成
CancelOrderRejectReason_UnknownOrder = 102  # 未知委托
CancelOrderRejectReason_BrokerOption = 103  # 柜台设置
CancelOrderRejectReason_AlreadyInPendingCancel = 104  # 委托撤销中

OrderSide_Unknown = 0
OrderSide_Buy = 1  # 买入
OrderSide_Sell = 2  # 卖出

OrderType_Unknown = 0
OrderType_Limit = 1  # 限价委托
OrderType_Market = 2  # 市价委托
OrderType_Stop = 3  # 止损止盈委托

OrderDuration_Unknown = 0
OrderDuration_FAK = 1  # 即时成交剩余撤销(fill and kill)
OrderDuration_FOK = 2  # 即时全额成交或撤销(fill or kill)
OrderDuration_GFD = 3  # 当日有效(good for day)
OrderDuration_GFS = 4  # 本节有效(good for section)
OrderDuration_GTD = 5  # 指定日期前有效(goodltilldate)
OrderDuration_GTC = 6  # 撤销前有效(goodtillcancel)
OrderDuration_GFA = 7  # 集合竞价前有效(good for auction)

OrderQualifier_Unknown = 0
OrderQualifier_BOC = 1  # 对方最优价格(best of counterparty)
OrderQualifier_BOP = 2  # 己方最优价格(best of party)
OrderQualifier_B5TC = 3  # 最优五档剩余撤销(best 5 then cancel)
OrderQualifier_B5TL = 4  # 最优五档剩余转限价(best 5 then limit)

OrderStyle_Unknown = 0
OrderStyle_Volume = 1
OrderStyle_Value = 2
OrderStyle_Percent = 3
OrderStyle_TargetVolume = 4
OrderStyle_TargetValue = 5
OrderStyle_TargetPercent = 6

PositionSide_Unknown = 0
PositionSide_Long = 1  # 多方向
PositionSide_Short = 2  # 空方向

PositionEffect_Unknown = 0
PositionEffect_Open = 1  # 开仓
PositionEffect_Close = 2  # 平仓, 具体语义取决于对应的交易所
PositionEffect_CloseToday = 3  # 平今仓
PositionEffect_CloseYesterday = 4  # 平昨仓

CashPositionChangeReason_Unknown = 0
CashPositionChangeReason_Trade = 1  # 交易
CashPositionChangeReason_Inout = 2  # 出入金 / 出入持仓

MODE_UNKNOWN = 0
MODE_LIVE = 1
MODE_BACKTEST = 2

ADJUST_NONE = 0
ADJUST_PREV = 1
ADJUST_POST = 2


SEC_TYPE_STOCK=1
SEC_TYPE_BOND=2
SEC_TYPE_FUND=3
SEC_TYPE_WARRANT=4
SEC_TYPE_INDEX=5
SEC_TYPE_FUTURES=6
SEC_TYPE_STOCK_KC=7
SEC_TYPE_PLATE=20
SEC_TYPE_CONCEPT=21






# coding=utf-8
from __future__ import unicode_literals, print_function, absolute_import

SUB_ID = 'SUB_ID:{}:{}:{}'  # 订阅股票唯一id  由三部分组成  symbol frequency count
SUB_TAG = 'SUB_TAG:{}:{}'  # 取消订阅股票标签 由两部分组成  symbol frequency

DATA_TYPE_TICK = 'tick'
DATA_TYPE_BAR = 'bar'

"""
c-sdk回调的类型
"""
CALLBACK_TYPE_INIT = 'init'
CALLBACK_TYPE_TICK = 'data.api.Tick'
CALLBACK_TYPE_BAR = 'data.api.Bar'
CALLBACK_TYPE_SCHEDULE = 'schedule'
CALLBACK_TYPE_EXECRPT = 'core.api.ExecRpt'
CALLBACK_TYPE_ORDER = 'core.api.Order'
CALLBACK_TYPE_INDICATOR = 'core.api.Indicator'
CALLBACK_TYPE_CASH = 'core.api.Cash'
CALLBACK_TYPE_POSITION = 'core.api.Position'
CALLBACK_TYPE_PARAMETERS = 'runtime-config'
CALLBACK_TYPE_ERROR = 'error'
CALLBACK_TYPE_TIMER = 'timer'
CALLBACK_TYPE_BACKTEST_FINISH = 'backtest-finished'
CALLBACK_TYPE_STOP = 'stop'

CALLBACK_TYPE_TRADE_CONNECTED = 'td-connected'
CALLBACK_TYPE_TRADE_DISCONNECTED = 'td-disconnected'

CALLBACK_TYPE_DATA_CONNECTED = 'md-connected'
CALLBACK_TYPE_DATA_DISCONNECTED = 'md-disconnected'

CALLBACK_TYPE_ACCOUNTSTATUS = 'core.api.AccountStatus'

TRADE_CONNECTED = 1
DATA_CONNECTED = 2

SCHEDULE_INFO = 'date_rule={date_rule},time_rule={time_rule}'

HISTORY_ADDR = 'bw-history-rpc'
HISTORY_REST_ADDR = 'bw-history-rpcgw'
FUNDAMENTAL_ADDR = 'bw-fundamental-rpc'

CSDK_OPERATE_SUCCESS = 0  # c-sdk 操作成功



class QuoteItemLikeDict2(dict):
    bid_p = ... # type: float
    bid_v = ... # type: int
    ask_p = ... # type: float
    ask_v = ... # type: int


class ObjectLikeDict(dict):
    quotes = ... # type: List[QuoteItemLikeDict2]
    symbol = ... # type: Text
    created_at = ... # type: datetime.datetime
    price = ... # type: float
    open = ... # type: float
    high = ... # type: float
    low = ... # type: float
    cum_volume = ... # type: float
    cum_amount = ... # type: float
    cum_position = ... # type: int
    last_amount = ... # type: float
    last_volume = ... # type: int
    trade_type = ... # type: int
    nanos = ... # type: int
    receive_local_time = ... # type: float


class BarLikeDict2(dict):
    symbol = ... # type: Text
    eob = ... # type: datetime.datetime
    bob = ... # type: datetime.datetime
    open = ... # type: float
    close = ... # type: float
    high = ... # type: float
    low = ... # type: float
    volume = ... # type: float
    amount = ... # type: float
    pre_close = ... # type: float
    position = ... # type: int
    frequency = ... # type: Text
    receive_local_time = ... # type: float




from typing import Text, Dict

# # nipasbc
# class GetFundamentalsReq(object):
#     table = ...  # type: Text
#     fields = ... # type: Text
#     filter = ... # type: Text
#     symbols = ... # type: Text
#     start_date = ... # type: Text
#     end_date = ... # type: Text
#     order_by = ... # type: Text
#     limit = ... # type: int

#     def __init__(self, table:Text, symbols:Text, end_date:Text, start_date:Text='', fields:Text='',  filter:Text='', order_by:Text='', limit:int=None): ...


# # nipasbc
# class GetFundamentalsNReq(object):
#     table = ...  # type: Text
#     fields = ... # type: Text
#     filter = ... # type: Text
#     symbols = ... # type: Text
#     end_date = ... # type: Text
#     order_by = ... # type: Text
#     count = ... # type: int

#     def __init__(self, table:Text, symbols:Text, end_date:Text, count:int, fields:Text='', filter:Text='', order_by:Text=''): ...

# # nipasbc
# class GetFundamentalsRsp(object):
#     class Fundamental(object):
#         symbol = ... # type: Text
#         fields = ... # type: Dict[Text, float]
#         pub_date = ... # type: int
#         end_date = ... # type: int

#         # def __init__(self, symbol:Text='', fields:Dict[Text, float]=None, pub_date:int=None, end_date:int=None): ...

#     data = ... # type: List[Fundamental]
#     # def __init__(self, data: List[Fundamental]=None):...


# # nipasbc
# class GetInstrumentInfosReq(object):
#     symbols = ... # type: Text
#     exchanges = ... # type: Text
#     sec_types = ... # type: Text
#     names = ... # type: Text
#     fields = ... # type: Text

#     def __init__(self, symbols:Text='', exchanges:Text='', sec_types:Text='', names:Text='', fields:Text=''): ...

# # nipasbc
# class GetInstrumentsReq(object):
#     symbols = ... # type:Text
#     exchanges = ... # type:Text
#     sec_types = ... # type:Text
#     names = ... # type:Text
#     skip_suspended = ...  # type: bool
#     skip_st = ...  # type: bool
#     fields = ...  # type: Text

#     def __init__(self, symbols:Text='', exchanges:Text='', sec_types:Text='', names:Text='', skip_suspended:bool=False,
#                  skip_st:bool=False, fields:Text=''): ...

# # nipasbc
# class GetHistoryInstrumentsReq(object):
#     symbols = ... # type: Text
#     fields = ... # type: Text
#     start_date = ... # type: Text
#     end_date = ... # type: Text

#     def __init__(self, symbols:Text='', fields:Text='', start_date:Text='', end_date:Text=''):...

# # nipasbc
# class GetConstituentsReq(object):
#     index = ...  # type: Text
#     fields = ...  # type: Text
#     start_date = ...  # type: Text
#     end_date = ...  # type: Text

#     def __init__(self, index:Text='', fields:Text='', start_date:Text='', end_date:Text=''): ...

# # nipasbc
# class GetSectorReq(object):
#     code = ...  # type: Text
#     def __init__(self, code:Text=''): ...

# # nipasbc
# class GetSectorRsp(object):
#     symbols = ... # type: List[Text]
#     # def __init__(self, symbols:List[Text]=None): ...

# # nipasbc
# class GetIndustryReq(object):
#     code = ...  # type: Text
#     def __init__(self, code:Text=''): ...

# # nipasbc
# class GetIndustryRsp(object):
#     symbols = ... # type: List[Text]
#     # def __init__(self, symbols:List[Text]=None): ...

# # nipasbc
# class GetConceptReq(object):
#     code = ...  # type: Text
#     def __init__(self, code:Text=''): ...

# # nipasbc
# class GetConceptRsp(object):
#     symbols = ... # type: List[Text]
#     # def __init__(self, symbols:List[Text]=None): ...

# # nipasbc
# class GetTradingDatesReq(object):
#     exchange = ...  # type: Text
#     start_date = ...  # type: Text
#     end_date = ...  # type: Text
#     def __init__(self, exchange:Text='', start_date:Text='', end_date:Text=''): ...

# # nipasbc
# class GetTradingDatesRsp(object):
#     dates = ... # type: List[int]
#     # def __init__(self, dates:List[int]=None): ...

# # nipasbc
# class GetPreviousTradingDateReq(object):
#     exchange = ...  # type: Text
#     date = ...  # type: Text
#     def __init__(self, exchange:Text='', date:Text=''): ...

# # nipasbc
# class GetPreviousTradingDateRsp(object):
#     date = ... # type: int
#     # def __init__(self, date:int=None): ...

# # nipasbc
# class GetNextTradingDateReq(object):
#     exchange = ...  # type: Text
#     date = ...  # type: Text
#     def __init__(self, exchange:Text='', date:Text=''): ...

# # nipasbc
# class GetNextTradingDateRsp(object):
#     date = ... # type: int
#     # def __init__(self, date:int=None): ...

# # nipasbc
# class GetDividendsReq(object):
#     symbol = ...  # type: Text
#     start_date = ...  # type: Text
#     end_date = ...  # type: Text
#     def __init__(self, symbol:Text='', start_date:Text='', end_date:Text=''): ...

# # nipasbc
# class GetDividendsSnapshotReq(object):
#     symbols = ...  # type: List[Text]
#     date = ...  # type: Text
#     # def __init__(self, symbols:List[Text]=None, date:Text=''): ...

# # nipasbc
# class GetContinuousContractsReq(object):
#     csymbol = ... # type: Text
#     start_date = ... # type: Text
#     end_date = ... # type: Text

#     def __init__(self, csymbol:Text='', start_date:Text='', end_date:Text=''):...



# coding=utf-8
from __future__ import unicode_literals, print_function, absolute_import

from datetime import datetime

from typing import NoReturn, Text, Optional




class DictLikeExecRpt(dict):
    """
    A dict that allows for ExecRpt-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeExecRpt, self).__str__()

    def __repr__(self):
        return super(DictLikeExecRpt, self).__repr__()

    @property
    def strategy_id(self):
        # type: ()->Text
        """
        策略ID
        """
        return self['strategy_id']

    @strategy_id.setter
    def strategy_id(self, value):
        # type: (Text) -> NoReturn
        self['strategy_id'] = value

    @property
    def account_id(self):
        # type: ()->Text
        """
        账号ID
        """
        return self['account_id']

    @account_id.setter
    def account_id(self, value):
        # type: (Text) -> NoReturn
        self['account_id'] = value

    @property
    def account_name(self):
        # type: ()->Text
        """
        账户登录名
        """
        return self['account_name']

    @account_name.setter
    def account_name(self, value):
        # type: (Text) -> NoReturn
        self['account_name'] = value

    @property
    def cl_ord_id(self):
        # type: ()->Text
        """
        委托客户端ID
        """
        return self['cl_ord_id']

    @cl_ord_id.setter
    def cl_ord_id(self, value):
        # type: (Text) -> NoReturn
        self['cl_ord_id'] = value

    @property
    def order_id(self):
        # type: ()->Text
        """
        委托柜台ID
        """
        return self['order_id']

    @order_id.setter
    def order_id(self, value):
        # type: (Text) -> NoReturn
        self['order_id'] = value

    @property
    def exec_id(self):
        # type: ()->Text
        """
        委托回报ID
        """
        return self['exec_id']

    @exec_id.setter
    def exec_id(self, value):
        # type: (Text) -> NoReturn
        self['exec_id'] = value

    @property
    def symbol(self):
        # type: ()->Text
        """
        symbol
        """
        return self['symbol']

    @symbol.setter
    def symbol(self, value):
        # type: (Text) -> NoReturn
        self['symbol'] = value

    @property
    def position_effect(self):
        # type: ()->int
        """
        开平标志，取值参考enum PositionEffect
        """
        return self['position_effect']

    @position_effect.setter
    def position_effect(self, value):
        # type: (int) -> NoReturn
        self['position_effect'] = value

    @property
    def side(self):
        # type: ()->int
        """
        买卖方向，取值参考enum OrderSide
        """
        return self['side']

    @side.setter
    def side(self, value):
        # type: (int) -> NoReturn
        self['side'] = value

    @property
    def ord_rej_reason(self):
        # type: ()->int
        """
        委托拒绝原因，取值参考enum OrderRejectReason
        """
        return self['ord_rej_reason']

    @ord_rej_reason.setter
    def ord_rej_reason(self, value):
        # type: (int) -> NoReturn
        self['ord_rej_reason'] = value

    @property
    def ord_rej_reason_detail(self):
        # type: ()->Text
        """
        委托拒绝原因描述
        """
        return self['ord_rej_reason_detail']

    @ord_rej_reason_detail.setter
    def ord_rej_reason_detail(self, value):
        # type: (Text) -> NoReturn
        self['ord_rej_reason_detail'] = value

    @property
    def exec_type(self):
        # type: ()->Text
        """
        执行回报类型, 取值参考enum ExecType
        """
        return self['exec_type']

    @exec_type.setter
    def exec_type(self, value):
        # type: (Text) -> NoReturn
        self['exec_type'] = value

    @property
    def price(self):
        # type: ()->float
        """
        委托成交价格
        """
        return self['price']

    @price.setter
    def price(self, value):
        # type: (float) -> NoReturn
        self['price'] = value

    @property
    def volume(self):
        # type: ()->int
        """
        委托成交量
        """
        return self['volume']

    @volume.setter
    def volume(self, value):
        # type: (int) -> NoReturn
        self['volume'] = value

    @property
    def amount(self):
        # type: ()->float
        """
        委托成交金额
        """
        return self['amount']

    @amount.setter
    def amount(self, value):
        # type: (float) -> NoReturn
        self['amount'] = value

    @property
    def commission(self):
        # type: ()->Text
        """
        委托成交手续费
        """
        return self['commission']

    @commission.setter
    def commission(self, value):
        # type: (Text) -> NoReturn
        self['commission'] = value

    @property
    def const(self):
        # type: ()->float
        """
        成本
        """
        return self['const']

    @const.setter
    def const(self, value):
        # type: (float) -> NoReturn
        self['const'] = value

    @property
    def created_at(self):
        # type: ()->Optional[datetime]
        """
        回报创建时间
        """
        return self['created_at']

    @created_at.setter
    def created_at(self, value):
        # type: (Optional[datetime]) -> NoReturn
        self['created_at'] = value


class DictLikeOrder(dict):
    """
    A dict that allows for Order-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeOrder, self).__str__()

    def __repr__(self):
        return super(DictLikeOrder, self).__repr__()

    @property
    def strategy_id(self):
        # type: ()->Text
        """
        策略ID
        """
        return self['strategy_id']

    @strategy_id.setter
    def strategy_id(self, value):
        # type: (Text) -> NoReturn
        self['strategy_id'] = value

    @property
    def account_id(self):
        # type: ()->Text
        """
        账号ID
        """
        return self['account_id']

    @account_id.setter
    def account_id(self, value):
        # type: (Text) -> NoReturn
        self['account_id'] = value

    @property
    def account_name(self):
        # type: ()->Text
        """
        账户登录名
        """
        return self['account_name']

    @account_name.setter
    def account_name(self, value):
        # type: (Text) -> NoReturn
        self['account_name'] = value

    @property
    def cl_ord_id(self):
        # type: ()->Text
        """
        委托客户端ID
        """
        return self['cl_ord_id']

    @cl_ord_id.setter
    def cl_ord_id(self, value):
        # type: (Text) -> NoReturn
        self['cl_ord_id'] = value

    @property
    def order_id(self):
        # type: ()->Text
        """
        委托柜台ID
        """
        return self['order_id']

    @order_id.setter
    def order_id(self, value):
        # type: (Text) -> NoReturn
        self['order_id'] = value

    @property
    def ex_ord_id(self):
        # type: ()->Text
        """
        委托交易所ID
        """
        return self['ex_ord_id']

    @ex_ord_id.setter
    def ex_ord_id(self, value):
        # type: (Text) -> NoReturn
        self['ex_ord_id'] = value

    @property
    def symbol(self):
        # type: ()->Text
        """
        symbol
        """
        return self['symbol']

    @symbol.setter
    def symbol(self, value):
        # type: (Text) -> NoReturn
        self['symbol'] = value

    @property
    def side(self):
        # type: ()->int
        """
        买卖方向，取值参考enum OrderSide
        """
        return self['side']

    @side.setter
    def side(self, value):
        # type: (int) -> NoReturn
        self['side'] = value

    @property
    def position_effect(self):
        # type: ()->int
        """
        开平标志，取值参考enum PositionEffect
        """
        return self['position_effect']

    @position_effect.setter
    def position_effect(self, value):
        # type: (int) -> NoReturn
        self['position_effect'] = value

    @property
    def position_side(self):
        # type: ()->int
        """
        持仓方向，取值参考enum PositionSide
        """
        return self['position_side']

    @position_side.setter
    def position_side(self, value):
        # type: (int) -> NoReturn
        self['position_side'] = value

    @property
    def order_type(self):
        # type: ()->int
        """
        委托类型，取值参考enum OrderType
        """
        return self['order_type']

    @order_type.setter
    def order_type(self, value):
        # type: (int) -> NoReturn
        self['order_type'] = value

    @property
    def order_duration(self):
        # type: ()->int
        """
        委托时间属性，取值参考enum OrderDuration
        """
        return self['order_duration']

    @order_duration.setter
    def order_duration(self, value):
        # type: (int) -> NoReturn
        self['order_duration'] = value

    @property
    def order_qualifier(self):
        # type: ()->int
        """
        委托成交属性，取值参考enum OrderQualifier
        """
        return self['order_qualifier']

    @order_qualifier.setter
    def order_qualifier(self, value):
        # type: (int) -> NoReturn
        self['order_qualifier'] = value

    @property
    def order_src(self):
        # type: ()->int
        """
        委托来源，取值参考enum OrderSrc
        """
        return self['order_src']

    @order_src.setter
    def order_src(self, value):
        # type: (int) -> NoReturn
        self['order_src'] = value

    @property
    def status(self):
        # type: ()->int
        """
        委托状态，取值参考enum OrderStatus
        """
        return self['status']

    @status.setter
    def status(self, value):
        # type: (int) -> NoReturn
        self['status'] = value

    @property
    def ord_rej_reason(self):
        # type: ()->int
        """
        委托拒绝原因，取值参考enum OrderRejectReason
        """
        return self['ord_rej_reason']

    @ord_rej_reason.setter
    def ord_rej_reason(self, value):
        # type: (int) -> NoReturn
        self['ord_rej_reason'] = value

    @property
    def ord_rej_reason_detail(self):
        # type: ()->int
        """
        委托拒绝原因描述
        """
        return self['ord_rej_reason_detail']

    @ord_rej_reason_detail.setter
    def ord_rej_reason_detail(self, value):
        # type: (int) -> NoReturn
        self['ord_rej_reason_detail'] = value

    @property
    def price(self):
        # type: ()->float
        """
        委托价格
        """
        return self['price']

    @price.setter
    def price(self, value):
        # type: (float) -> NoReturn
        self['price'] = value

    @property
    def stop_price(self):
        # type: ()->float
        """
        委托止损/止盈触发价格
        """
        return self['stop_price']

    @stop_price.setter
    def stop_price(self, value):
        # type: (float) -> NoReturn
        self['stop_price'] = value

    @property
    def order_style(self):
        # type: ()->int
        """
        委托风格，取值参考 enum OrderStyle
        """
        return self['order_style']

    @order_style.setter
    def order_style(self, value):
        # type: (int) -> NoReturn
        self['order_style'] = value

    @property
    def volume(self):
        # type: ()->int
        """
        委托量
        """
        return self['volume']

    @volume.setter
    def volume(self, value):
        # type: (int) -> NoReturn
        self['volume'] = value

    @property
    def value(self):
        # type: ()->float
        """
        委托额
        """
        return self['value']

    @value.setter
    def value(self, value):
        # type: (float) -> NoReturn
        self['value'] = value

    @property
    def percent(self):
        # type: ()->float
        """
        委托百分比
        """
        return self['percent']

    @percent.setter
    def percent(self, value):
        # type: (float) -> NoReturn
        self['percent'] = value

    @property
    def target_volume(self):
        # type: ()->int
        """
        委托目标量
        """
        return self['target_volume']

    @target_volume.setter
    def target_volume(self, value):
        # type: (int) -> NoReturn
        self['target_volume'] = value

    @property
    def target_value(self):
        # type: ()->float
        """
        委托目标额
        """
        return self['target_value']

    @target_value.setter
    def target_value(self, value):
        # type: (float) -> NoReturn
        self['target_value'] = value

    @property
    def target_percent(self):
        # type: ()->float
        """
        委托目标百分比
        """
        return self['target_percent']

    @target_percent.setter
    def target_percent(self, value):
        # type: (float) -> NoReturn
        self['target_percent'] = value

    @property
    def filled_volume(self):
        # type: ()->int
        """
        已成量
        """
        return self['filled_volume']

    @filled_volume.setter
    def filled_volume(self, value):
        # type: (int) -> NoReturn
        self['filled_volume'] = value

    @property
    def filled_vwap(self):
        # type: ()->float
        """
        已成均价
        """
        return self['filled_vwap']

    @filled_vwap.setter
    def filled_vwap(self, value):
        # type: (float) -> NoReturn
        self['filled_vwap'] = value

    @property
    def filled_amount(self):
        # type: ()->float
        """
        已成金额
        """
        return self['filled_amount']

    @filled_amount.setter
    def filled_amount(self, value):
        # type: (float) -> NoReturn
        self['filled_amount'] = value

    @property
    def filled_commission(self):
        # type: ()->float
        """
        已成手续费
        """
        return self['filled_commission']

    @filled_commission.setter
    def filled_commission(self, value):
        # type: (float) -> NoReturn
        self['filled_commission'] = value

    @property
    def created_at(self):
        # type: ()->Optional[datetime]
        """
        委托创建时间
        """
        return self['created_at']

    @created_at.setter
    def created_at(self, value):
        # type: (Optional[datetime]) -> NoReturn
        self['created_at'] = value

    @property
    def updated_at(self):
        # type: ()->Optional[datetime]
        """
        委托更新时间
        """
        return self['updated_at']

    @updated_at.setter
    def updated_at(self, value):
        # type: (Optional[datetime]) -> NoReturn
        self['updated_at'] = value


class DictLikeIndicator(dict):
    """
    A dict that allows for Indicator-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeIndicator, self).__str__()

    def __repr__(self):
        return super(DictLikeIndicator, self).__repr__()

    @property
    def account_id(self):
        # type: ()->Text
        """
        账号ID
        """
        return self['account_id']

    @account_id.setter
    def account_id(self, value):
        # type: (Text) -> NoReturn
        self['account_id'] = value

    @property
    def pnl_ratio(self):
        # type: ()->float
        """
        累计收益率(pnl/cum_inout)
        """
        return self['pnl_ratio']

    @pnl_ratio.setter
    def pnl_ratio(self, value):
        # type: (float) -> NoReturn
        self['pnl_ratio'] = value

    @property
    def pnl_ratio_annual(self):
        # type: ()->float
        """
        年化收益率
        """
        return self['pnl_ratio_annual']

    @pnl_ratio_annual.setter
    def pnl_ratio_annual(self, value):
        # type: (float) -> NoReturn
        self['pnl_ratio_annual'] = value

    @property
    def sharp_ratio(self):
        # type: ()->float
        """
        夏普比率
        """
        return self['sharp_ratio']

    @sharp_ratio.setter
    def sharp_ratio(self, value):
        # type: (float) -> NoReturn
        self['sharp_ratio'] = value

    @property
    def max_drawdown(self):
        # type: ()->float
        """
        最大回撤
        """
        return self['max_drawdown']

    @max_drawdown.setter
    def max_drawdown(self, value):
        # type: (float) -> NoReturn
        self['max_drawdown'] = value

    @property
    def risk_ratio(self):
        # type: ()->float
        """
        风险比率
        """
        return self['risk_ratio']

    @risk_ratio.setter
    def risk_ratio(self, value):
        # type: (float) -> NoReturn
        self['risk_ratio'] = value

    @property
    def open_count(self):
        # type: ()->int
        """
        开仓次数
        """
        return self['open_count']

    @open_count.setter
    def open_count(self, value):
        # type: (int) -> NoReturn
        self['open_count'] = value

    @property
    def close_count(self):
        # type: ()->int
        """
        平仓次数
        """
        return self['close_count']

    @close_count.setter
    def close_count(self, value):
        # type: (int) -> NoReturn
        self['close_count'] = value

    @property
    def win_count(self):
        # type: ()->int
        """
        盈利次数
        """
        return self['win_count']

    @win_count.setter
    def win_count(self, value):
        # type: (int) -> NoReturn
        self['win_count'] = value

    @property
    def lose_count(self):
        # type: ()->int
        """
        亏损次数
        """
        return self['lose_count']

    @lose_count.setter
    def lose_count(self, value):
        # type: (int) -> NoReturn
        self['lose_count'] = value

    @property
    def win_ratio(self):
        # type: ()->float
        """
        胜率
        """
        return self['win_ratio']

    @win_ratio.setter
    def win_ratio(self, value):
        # type: (float) -> NoReturn
        self['win_ratio'] = value

    @property
    def calmar_ratio(self):
        # type: ()->float
        """
        卡玛比率
        """
        return self['calmar_ratio']

    @calmar_ratio.setter
    def calmar_ratio(self, value):
        # type: (float) -> NoReturn
        self['calmar_ratio'] = value

    @property
    def created_at(self):
        # type: ()->Optional[datetime]
        """

        """
        return self['created_at']

    @created_at.setter
    def created_at(self, value):
        # type: (Optional[datetime]) -> NoReturn
        self['created_at'] = value

    @property
    def updated_at(self):
        # type: ()->Optional[datetime]
        """

        """
        return self['updated_at']

    @updated_at.setter
    def updated_at(self, value):
        # type: (Optional[datetime]) -> NoReturn
        self['updated_at'] = value


class DictLikeParameter(dict):
    """
    A dict that allows for Parameter-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeParameter, self).__str__()

    def __repr__(self):
        return super(DictLikeParameter, self).__repr__()

    @property
    def key(self):
        # type: ()->Text
        """

        """
        return self['key']

    @key.setter
    def key(self, value):
        # type: (Text) -> NoReturn
        self['key'] = value

    @property
    def value(self):
        # type: ()->float
        """

        """
        return self['value']

    @value.setter
    def value(self, value):
        # type: (float) -> NoReturn
        self['value'] = value

    @property
    def min(self):
        # type: ()->float
        """

        """
        return self['min']

    @min.setter
    def min(self, value):
        # type: (float) -> NoReturn
        self['min'] = value

    @property
    def max(self):
        # type: ()->float
        """

        """
        return self['max']

    @max.setter
    def max(self, value):
        # type: (float) -> NoReturn
        self['max'] = value

    @property
    def name(self):
        # type: ()->Text
        """

        """
        return self['name']

    @name.setter
    def name(self, value):
        # type: (Text) -> NoReturn
        self['name'] = value

    @property
    def intro(self):
        # type: ()->Text
        """

        """
        return self['intro']

    @intro.setter
    def intro(self, value):
        # type: (Text) -> NoReturn
        self['intro'] = value

    @property
    def group(self):
        # type: ()->Text
        """

        """
        return self['group']

    @group.setter
    def group(self, value):
        # type: (Text) -> NoReturn
        self['group'] = value

    @property
    def readonly(self):
        # type: ()->bool
        """

        """
        return self['readonly']

    @readonly.setter
    def readonly(self, value):
        # type: (bool) -> NoReturn
        self['readonly'] = value


class DictLikeAccountStatus(dict):
    """
    A dict that allows for AccountStatus-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeAccountStatus, self).__str__()

    def __repr__(self):
        return super(DictLikeAccountStatus, self).__repr__()

    @property
    def account_id(self):
        # type: ()->Text

        return self['account_id']

    @account_id.setter
    def account_id(self, value):
        # type: (Text) -> NoReturn
        self['account_id'] = value

    @property
    def account_name(self):
        # type: ()->Text

        return self['account_name']

    @account_name.setter
    def account_name(self, value):
        # type: (Text) -> NoReturn
        self['account_name'] = value

    @property
    def status(self):
        # type: ()->DictLikeConnectionStatus

        return self['status']

    @status.setter
    def status(self, value):
        # type: (DictLikeConnectionStatus) -> NoReturn
        self['status'] = value


class DictLikeConnectionStatus(dict):
    """
    A dict that allows for ConnectionStatus-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeConnectionStatus, self).__str__()

    def __repr__(self):
        return super(DictLikeConnectionStatus, self).__repr__()

    @property
    def state(self):
        # type: ()->int

        return self['state']

    @state.setter
    def state(self, value):
        # type: (int) -> NoReturn
        self['state'] = value

    @property
    def error(self):
        # type: ()->DictLikeError

        return self['error']

    @error.setter
    def error(self, value):
        # type: (DictLikeError) -> NoReturn
        self['error'] = value


class DictLikeError(dict):
    """
    A dict that allows for Error-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeError, self).__str__()

    def __repr__(self):
        return super(DictLikeError, self).__repr__()

    @property
    def code(self):
        # type: ()->int

        return self['code']

    @code.setter
    def code(self, value):
        # type: (int) -> NoReturn
        self['code'] = value

    @property
    def type(self):
        # type: ()->Text

        return self['type']

    @type.setter
    def type(self, value):
        # type: (Text) -> NoReturn
        self['type'] = value

    @property
    def info(self):
        # type: ()->Text

        return self['info']

    @info.setter
    def info(self, value):
        # type: (Text) -> NoReturn
        self['info'] = value



# coding=utf-8
from __future__ import unicode_literals, print_function, absolute_import

import logging
import sys
import time

import arrow
import six
from arrow import parser
from typing import Dict, Text, Any, List, Union
from datetime import date, datetime, timedelta, tzinfo
from dateutil.tz import tzutc

BwDate = Union[Text, datetime, date]  # 自定义bw里可表示时间的类型
TextNone = Union[Text, None]  # 可表示str或者None类型
BwSymbols = Union[Text, List[Text]]  # 用逗号分割或list股票代码表示多个股票

bwsdklogger = logging.getLogger('bwsdklogger')
bwsdklogger.setLevel(logging.INFO)
console_handler = logging.StreamHandler(stream=sys.stdout)
console_handler.setLevel(logging.DEBUG)
bwformat = logging.Formatter('%(asctime)-15s %(levelname)s [%(filename)s:%(lineno)d] %(message)s')
console_handler.setFormatter(bwformat)
bwsdklogger.addHandler(console_handler)

__add_logfile = False


def bwsdklog2file():
    """
    增加一个handler, 把bwsdklog写入到日志文件. 只调用一次
    """
    global __add_logfile
    if __add_logfile:
        return

    file_handler = logging.FileHandler('bwsdk_{}.log'.format(datetime.now().strftime("%Y%m%d%H%M%S")))
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(bwformat)
    bwsdklogger.addHandler(file_handler)
    __add_logfile = True


def str_lowerstrip(s):
    # type: (Text) -> Text
    """
    把字符串进行lower跟strip操作
    """
    return s.lower().strip() if s else ''


def date2datetime(d):
    # type: (Union[date,str]) -> datetime
    """
    把date类型转换为datetime类型
    """
    if isinstance(d, six.string_types):
        if len(d) == 8:
            return datetime.strptime(d, '%Y%m%d')
        if len(d) == 10:
            return datetime.strptime(d, '%Y-%m-%d')
        raise Exception('字符串{}不能转为datetime'.format(d))
    return datetime.combine(d, datetime.min.time())


def str2datetime(d):
    # type: (Text) -> Union[datetime, None]
    """
    把字符串转成datetime类型
    """
    if len(d) == 8:
        return datetime.strptime(d, '%Y%m%d')
    if len(d) == 10:
        return datetime.strptime(d, '%Y-%m-%d')
    if len(d) == 17:  # 类于 201707010 8:50:00
        return datetime.strptime(d, '%Y%m%d %H:%M:%S')
    if len(d) == 19:  # 类于 2017-07-01 08:50:00
        return datetime.strptime(d, '%Y-%m-%d %H:%M:%S')
    bwsdklogger.error("字符串 %s 不能格式化成datetime类型, 请使用类似于 2017-07-01 08:50:00 这种格式", d)
    return None


def timestamp2bj_datetime(timestamp):
    # type: (int) -> Union[datetime, None]
    """
    把 protobuf 的 timestamp(utc) 转为北京时间 datetime类型
    :return:
    """
    if timestamp is None:
        return None
    # +8 小时
    deltasec = timestamp.seconds + 8 * 60 * 60 + timestamp.nanos / float(_NANOS_PER_SECOND)
    d = datetime(1970, 1, 1) + timedelta(seconds=deltasec)
    d = d.replace(tzinfo=beijing_tzinfo)
    return d


def timestamp2datetime(timestamp):
    # type: (int) -> Union[datetime, None]
    """
    把 protobuf 的 timestamp 转为 datetime类型
    :return:
    """
    if timestamp is None:
        return None
    deltasec = timestamp.seconds + timestamp.nanos / float(_NANOS_PER_SECOND)
    return datetime(1970, 1, 1) + timedelta(seconds=deltasec)


# 来自于 google.protobuf.internal.well_nkown_types.py 里定义的变量
_NANOS_PER_SECOND = 1000000000


def to_datestr(d):
    # type: (BwDate) -> TextNone
    """
    把datetime.date或datetime.datetime, 或者 yyyy-mm-dd, yyyymmdd 表示日期的字符串统一转换为 yyyy-mm-dd的字符串
    如果不能转换返回None
    """
    date_str = ''
    if isinstance(d, (date, datetime)):
        date_str = d.strftime('%Y-%m-%d')
    if isinstance(d, six.string_types):
        if len(d) == 8:
            try:
                dt = datetime.strptime(d, '%Y%m%d')
                date_str = dt.strftime('%Y-%m-%d')
            except ValueError as e:
                pass
        if len(d) == 10:
            try:
                datetime.strptime(d, '%Y-%m-%d')
                date_str = d
            except ValueError as e:
                pass

    return date_str if date_str else d


beijing_tzinfo = parser.TzinfoParser.parse('Asia/Shanghai')  # type: tzinfo
utc_tzinfo = parser.TzinfoParser.parse('utc')  # type: tzinfo


def utc_datetime2beijing_datetime(dt):
    # type: (datetime) -> datetime
    """
    把utc的datetime转化为北京时间的datetime. 这样2016-04-16T16:00:00+00:00 转化后得到的date就是 2016-04-17
    """
    return arrow.Arrow.fromdatetime(dt).to(beijing_tzinfo).datetime


def beijing_datetime2utc_datetime(dt):
    # type: (datetime) -> datetime
    """
    把北京时间的datetime转化为utc的datetime
    """
    return arrow.Arrow.fromdatetime(dt, beijing_tzinfo).to(utc_tzinfo).datetime


def to_exchange(exchange):
    # type: (Text) -> TextNone
    """转换成正确的交易市场. 如果不存在, 则返回None"""
    exchange_set = {'SH', 'SZ', 'CFFEX', 'SHFE', 'DCE', 'CZCE', 'INE'}
    if not exchange:
        return None
    s = str(exchange).upper().strip()
    return s if s in exchange_set else None


def load_to_list(value):
    """
    无论输入的是啥类型， 都转成list
    """
    if isinstance(value, list):
        return value

    if isinstance(value, dict):
        return [value]

    if isinstance(value, six.string_types):
        if len(value)>0:
            return [item.strip() for item in value.split(',')]
        else:
            return []
    return value


def load_to_datetime_str(value):
    
    if isinstance(value, (date, datetime)):
        return value.strftime('%Y-%m-%d %H:%M:%S')

    return value


def to_datetime_bjtzinfo(value):
    if isinstance(value, datetime):
        pass
    if isinstance(value, date):
        value = datetime(value.year, value.month, value.day)
    if isinstance(value, six.string_types):
        try:
            value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        except Exception as identifier:
            # print("*"*100,identifier)
            try:
                value = datetime.strptime(value, '%Y-%m-%d')
            except Exception as identifier:
                value =datetime.strptime(value, '%Y%m%d')
            pass
        
    value = value.replace(tzinfo=beijing_tzinfo)
    return value


def load_to_second(value):
    # type: (Text) -> int
    value = value.strip().lower()

    if value.endswith('h'):
        return int(value[0:-1]) * 60 * 60

    if value.endswith('d'):
        return int(value[0:-1]) * 60 * 60 * 24

    if value.endswith('s'):
        return int(value[0:-1])

    raise ValueError('仅支持s(秒), h(小时), d(天) 结尾')


# def adjust_frequency(frequency):
#     # type: (Text)->Text
#     """
#     把frequency为1m这样的分钟, 调整为秒
#     """
#     frequency = frequency.strip().lower()
#     return '{}s'.format(int(frequency[0:-1] * 60)) if frequency.endswith('m') else frequency


class DictLikeObject(dict):
    """
    A dict that allows for object-like property access syntax.
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item, value):
        self[item] = value

    def __str__(self):
        return super(DictLikeObject, self).__str__()

    def __repr__(self):
        return super(DictLikeObject, self).__repr__()


class ObjectLikeDict(object):
    def __init__(self, data):
        self.data = data

    def __getitem__(self, item):
        return self.data.get(item)

    # 如果前面没有定义过的属性， 就会调用这个函数
    def __getattr__(self, item):
        return self.data.get(item)

    def __setitem__(self, item, value):
        self.data[item] = value

    def __setattr__(self, item, value):
        if item != 'data':
            self.data[item] = value
        else:
            super(ObjectLikeDict, self).__setattr__(item, value)

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.data.__repr__()


def datetime2timestamp(dt):
    """
    Converts a datetime object to UNIX timestamp in milliseconds.
    """
    if isinstance(dt, datetime):
        timestamp = time.mktime(dt.timetuple())
        return int(timestamp)
    return dt


def standard_fields(fields, letter_upper=False):

    if not fields:
        return None, None

    if isinstance(fields, list):
        fields= ','.join(fields)

    if letter_upper:
        fields_list = [field.strip().upper() for field in fields.split(',')]

    else:
        fields_list = [field.strip().lower() for field in fields.split(',')]

    fields_str = ','.join(fields_list)
    return fields_str, fields_list



# coding=utf-8
from __future__ import print_function, absolute_import, unicode_literals

import collections
import os
import sys
from importlib import import_module
from optparse import OptionParser

import six
from typing import Text, List, Dict, NoReturn, Any

import time


from datetime import date, datetime, timedelta, tzinfo

from bw.__version__ import __version__
from bw.callback import callback_controller
from bw.constant import DATA_TYPE_TICK, SCHEDULE_INFO

from bw.yy.data import Tick, Bar, Quote,Trade,Timestamp

from bw.csdk.cpp_sdk import py_bwi_current, py_bwi_schedule, py_bwi_set_data_callback, py_bwi_set_strategy_id, \
    py_bwi_set_mode, py_bwi_set_backtest_config, py_bwi_subscribe, py_bwi_set_token, py_bwi_unsubscribe, \
    py_bwi_log, py_bwi_strerror, py_bwi_set_user_token,py_bwi_set_pipe,py_bwi_set_srv_cache,\
    py_bwi_run, py_bwi_set_serv_addr, py_bwi_init, py_bwi_poll, py_bwi_get_c_version, py_bwi_set_apitoken,\
        py_bwi_set_cache

from bw.enum import MODE_UNKNOWN, ADJUST_NONE, MODE_BACKTEST, MODE_LIVE, ADJUST_POST, ADJUST_PREV
from bw.model.storage import context, BarSubInfo, BarWaitgroupInfo


from bw.utils import load_to_list, load_to_second, BwSymbols, bwsdklogger,\
timestamp2bj_datetime,timestamp2datetime

running = True


def _unsubscribe_all():
    context.bar_sub_infos.clear()
    context.tick_sub_symbols.clear()
    context.bar_waitgroup_frequency2Info.clear()
    context.bar_data_cache.clear()
    context.tick_data_cache.clear()
    context.max_tick_data_count = 1
    context.max_bar_data_count = 1


def set_token(usr,pwd):
    # type: (Text) ->NoReturn
    """
    设置用户的token，用于身份认证
    """
    py_bwi_set_token(usr,pwd)
    context.token = str('bearer {}'.format(usr))

def set_srv_cache(m):
    # type: (Text) ->NoReturn
    py_bwi_set_srv_cache(m)


def get_version():
    # type: () ->Text
    return __version__


def subscribe(symbols, frequency='1d', count=1, wait_group=False, wait_group_timeout='10s', unsubscribe_previous=False):
    # type:(BwSymbols, Text, int, bool, Text, bool) ->NoReturn
    """
    订阅行情，可以指定symbol， 数据滑窗大小，以及是否需要等待全部代码的数据到齐再触发事件。
    wait_group: 是否等待全部相同频度订阅的symbol到齐再触发on_bar事件。
    一个 frequency, 只能有一个 wait_group_timeout 也就是说如果后面调用该函数时, 
    相同的frequency, 但是 wait_group_timeout 不同,则 wait_group_timeout 被忽略.
    同时要注意, 一个symbol与frequency组合, 只能有一种wait_group, 即wait_group要么为true, 要么为false
    """
    # frequency = adjust_frequency(frequency)

    symbols = load_to_list(symbols)

    symbols_str = ','.join(symbols)
    py_bwi_subscribe(symbols_str, frequency, unsubscribe_previous)

    # print('*'*23,symbols_str,frequency)

    if unsubscribe_previous:
        _unsubscribe_all()

    if frequency == DATA_TYPE_TICK:
        if context.max_tick_data_count < count:
            context.max_tick_data_count = count
        for sy in symbols:
            context.tick_data_cache[sy] = collections.deque(maxlen=count)
            context.tick_sub_symbols.add(sy)
        return

    # 处理订阅bar的情况
    context._set_onbar_waitgroup_timeout_check()
    wait_group_timeoutint = load_to_second(wait_group_timeout)
    if context.max_bar_data_count < count:
        context.max_bar_data_count = count
    for sy in symbols:
        context.bar_data_cache["{}_{}".format(sy, frequency)] = collections.deque(maxlen=count)
        barsubinfo = BarSubInfo(sy, frequency)
        if barsubinfo not in context.bar_sub_infos:
            context.bar_sub_infos.add(barsubinfo)
            if wait_group:
                if frequency not in context.bar_waitgroup_frequency2Info:
                    context.bar_waitgroup_frequency2Info[frequency] = BarWaitgroupInfo(frequency, wait_group_timeoutint)
                context.bar_waitgroup_frequency2Info[frequency].add_one_symbol(sy)
        else:
            bwsdklogger.debug("symbol=%s frequency=%s 已订阅过", sy, frequency)
            continue


def unsubscribe(symbols, frequency='1d'):
    # type: (BwSymbols, Text) ->NoReturn
    """
    unsubscribe - 取消行情订阅

    取消行情订阅，默认取消所有已订阅行情
    """
    symbols = load_to_list(symbols)
    symbols_str = ','.join(symbols)
    # frequency = adjust_frequency(frequency)

    status = py_bwi_unsubscribe(symbols_str, frequency)

    if symbols_str == '*':
        _unsubscribe_all()
        return

    if frequency == DATA_TYPE_TICK:
        for sy in symbols:
            if sy in list(six.iterkeys(context.tick_data_cache)):
                del context.tick_data_cache[sy]
                context.tick_sub_symbols.remove(sy)
        return

    # 处理bar的退订
    for sy in symbols:
        k = sy + "_" + frequency
        if k in list(six.iterkeys(context.bar_data_cache)):
            del context.bar_data_cache[k]
            context.bar_sub_infos.remove(BarSubInfo(sy, frequency))
            barwaitgroupinfo = context.bar_waitgroup_frequency2Info.get(frequency, None)
            if barwaitgroupinfo:
                barwaitgroupinfo.remove_one_symbol(sy)

    # 处理已全部退订的 frequency
    for frequency in list(six.iterkeys(context.bar_waitgroup_frequency2Info)):
        if len(context.bar_waitgroup_frequency2Info[frequency]) == 0:
            bwsdklogger.debug('frequency=%s 已全部取消订阅', frequency)
            del context.bar_waitgroup_frequency2Info[frequency]


def current(symbols, fields=''):
    # type: (BwSymbols, Text) -> List[Any]
    """
    查询当前行情快照，返回tick数据
    """
    symbols = load_to_list(symbols)
    fields = load_to_list(fields)

    symbols_str = ','.join(symbols)
    fields_str = ','.join(fields)
    ticklist =[]
    data = py_bwi_current(symbols_str)

    if data is None:
        return []
    else:
        for item in data:
            tick = Tick(item) 
            tick.FromFmt(item) # 格式化成timestamp
            
            # print("*"*70,'current:',tick.created_at)
            # print (timestamp2bj_datetime(Timestamp(tick.created_at)))

            ticknew = {
            'quotes': tick.quotes,
            'symbol': tick.symbol,
            'created_at': timestamp2bj_datetime(tick.created_at),
            'price': tick.price,
            'open': tick.open,
            'high': tick.high,
            'low': tick.low,
            'cum_volume': tick.cum_volume,
            'cum_amount': tick.cum_amount,
            'cum_position': tick.cum_position,
            'last_amount': tick.last_amount,
            'last_volume': tick.last_volume,
            'trade_type': tick.trade_type,
            'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            ticknn = {}
            # print("*"*21,'current:',fields)
            for k,v in ticknew.items():
                # print(k,v)
                if len(fields)>0:
                    if k in fields:
                        ticknn[k] = v
                else:
                    # print(fields)
                    ticknn[k] = v
                pass
            # print(ticknn)
            ticklist.append(ticknn)
    
    return ticklist


def get_strerror(error_code):
    # type: (int) -> Text
    return py_bwi_strerror(error_code)

def set_cache(cmode):
    return py_bwi_set_cache(cmode)

def schedule(schedule_func, date_rule, time_rule):
    # type: (Any, Text, Text) ->NoReturn
    """
    定时任务. 这里的schedule_func 要求只能有一个context参数的函数
    """
    from datetime import datetime
    d = datetime.strptime(time_rule, '%H:%M:%S')
    time_rule = d.strftime("%H:%M:%S")
    # print(time_rule)
    

    schemdule_info = SCHEDULE_INFO.format(date_rule=date_rule, time_rule=time_rule)
    context.inside_schedules[schemdule_info] = schedule_func
    # print(context.inside_schedules)
    py_bwi_schedule(date_rule, time_rule)


def _all_not_none(a, b):
    # type: (Any, Any) -> bool
    """
    全部都不为None
    """
    return a is not None and b is not None


def run(strategy_id='', filename='', mode=MODE_UNKNOWN, usr='',
        # pwd='',
        backtest_start_time='',
        backtest_end_time='',
        backtest_initial_cash=1000000,
        # backtest_transaction_ratio=1,
        backtest_commission_ratio=0,
        backtest_slippage_ratio=0,
        backtest_adjust=ADJUST_PREV,
        backtest_check_cache=1,
        serv_addr='',
        ptoken='',
        ppipe='',
        pticket=''):
    # type: (Text, Text, int, Text, Text, Text, float, float, float, float, int, int, Text) ->NoReturn
    """
    执行策略
    """
    # print('filename in first:',filename)
    parser = OptionParser()
    parser.add_option("--strategy_id", action="store",
                      dest="strategy_id",
                      default=strategy_id,
                      help="策略id")

    parser.add_option("--filename", action="store",
                      dest="filename",
                      default=filename,
                      help="策略文件名称")

    parser.add_option("--mode", action="store",
                      dest="mode",
                      default=mode,
                      help="策略模式选择")

    parser.add_option("--usr", action="store",
                      dest="usr",
                      default=usr,
                      help="用户usr")

    # parser.add_option("--pwd", action="store",
    #                   dest="pwd",
    #                   default=pwd,
    #                   help="用户pwd")

    parser.add_option("--backtest_start_time", action="store",
                      dest="backtest_start_time",
                      default=backtest_start_time,
                      help="回测开始时间")

    parser.add_option("--backtest_end_time", action="store",
                      dest="backtest_end_time",
                      default=backtest_end_time,
                      help="回测结束时间")

    parser.add_option("--backtest_initial_cash", action="store",
                      dest="backtest_initial_cash",
                      default=backtest_initial_cash,
                      help="回测初始资金")

    # parser.add_option("--backtest_transaction_ratio", action="store",
    #                   dest="backtest_transaction_ratio",
    #                   default=backtest_transaction_ratio,
    #                   help="回测交易费率")

    parser.add_option("--backtest_commission_ratio", action="store",
                      dest="backtest_commission_ratio",
                      default=backtest_commission_ratio,
                      help="回测成交比率")

    parser.add_option("--backtest_slippage_ratio", action="store",
                      dest="backtest_slippage_ratio",
                      default=backtest_slippage_ratio,
                      help="回测滑点费率--暂无")

    parser.add_option("--backtest_adjust", action="store",
                      dest="backtest_adjust",
                      default=backtest_adjust,
                      help="回测复权模式")

    parser.add_option("--backtest_check_cache", action="store",
                      dest="backtest_check_cache",
                      default=backtest_check_cache,
                      help="回测是否使用缓存")

    parser.add_option("--serv_addr", action="store",
                      dest="serv_addr",
                      default=serv_addr,
                      help="终端地址")
    
    parser.add_option("--server", action="store",
                      dest="server",
                      default=serv_addr,
                      help="终端地址")
    
    parser.add_option("--token", action="store",
                      dest="token",
                      default=ptoken,
                      help="verifytoken accountid")

    parser.add_option("--pipe", action="store",
                      dest="pipe",
                      default=ppipe,
                      help="servnet")

    parser.add_option("--ticket", action="store",
                      dest="ticket",
                      default=pticket,
                      help="usertoken")

    (options, args) = parser.parse_args()
    # print('%'*50,options,'%'*50)
    strategy_id = options.strategy_id
    pipe = options.pipe
    tickx = options.ticket
    filename = options.filename
    mode = int(options.mode)
    if mode not in (MODE_UNKNOWN, MODE_LIVE, MODE_BACKTEST):
        raise ValueError('模式只能设置成 MODE_UNKNOWN, MODE_LIVE, MODE_BACKTEST 值')

    from decimal import Decimal
    usr = options.usr
    # pwd = options.pwd
    token=options.token
    
    backtest_start_time = options.backtest_start_time
    backtest_end_time = options.backtest_end_time
    backtest_initial_cash = float(options.backtest_initial_cash)
    # backtest_transaction_ratio = float(options.backtest_transaction_ratio)
    backtest_commission_ratio = float(options.backtest_commission_ratio)
    backtest_slippage_ratio = float(options.backtest_slippage_ratio)
    backtest_adjust = int(options.backtest_adjust)
    # print('*'*50)
    # print (backtest_start_time,backtest_end_time,backtest_initial_cash\
    #     ,backtest_commission_ratio,backtest_slippage_ratio,backtest_adjust)
    # print('*'*50)
    if backtest_adjust == 3:
        backtest_adjust = ADJUST_NONE  # 这个修改是为了适合终端之前把 3 认为是 不复权
    if backtest_adjust not in (ADJUST_NONE, ADJUST_POST, ADJUST_PREV):
        raise ValueError('回测复权模式只能设置成 ADJUST_NONE, ADJUST_POST, ADJUST_PREV 值')

    if backtest_initial_cash < 1:
        raise ValueError('回测初始资金不能设置为小于1, 当前值为:{}'.format(backtest_initial_cash))

    # if not 0 <= backtest_transaction_ratio <= 1:
    #     raise ValueError('回测成交比例允许的范围值为 0<=x<=1, 当前值为{}'.format(backtest_transaction_ratio))

    if not 0 <= backtest_commission_ratio <= 0.1:
        raise ValueError('回测手续费比例允许的范围值为 0<=x<=0.1, 当前值为{}'.format(backtest_commission_ratio))

    if not 0 <= backtest_slippage_ratio <= 0.1:
        raise ValueError('回测滑点比例允许的范围值为 0<=x<=0.1, 当前值为{}'.format(backtest_slippage_ratio))

    backtest_check_cache = int(options.backtest_check_cache)
    serv_addr = options.serv_addr
    serv_addr = options.server

    from bw import api

    # 处理用户传入 __file__这个特殊变量的情况
    syspathes = set(s.replace('\\', '/') for s in sys.path)
    commonpaths = [os.path.commonprefix([p, filename]) for p in syspathes]
    commonpaths.sort(key=lambda s: len(s), reverse=True)
    maxcommonpath = commonpaths[0]
    filename = filename.replace(maxcommonpath, '')  # type: str
    if filename.startswith('/'):
        filename = filename[1:]
    if filename.endswith(".py"):
        filename = filename[:-3]
    
    filename = filename.replace("/", ".")
    filename = filename.replace('\\', ".")
    
    fmodule = import_module(filename)

    # print('importmodule ',filename)
    # 把bw.api里的所有的符号都导出到当前策略文件(fmodule)的命令空间, 方便使用
    for name in api.__all__:
        if name not in fmodule.__dict__:
            fmodule.__dict__[name] = getattr(api, name)

    # 服务地址设置
    if serv_addr:
        set_serv_addr(serv_addr)

    set_token(usr,tickx)
    
    py_bwi_set_user_token(token)

    py_bwi_set_pipe(pipe)

    py_bwi_set_serv_addr(serv_addr)

    py_bwi_set_strategy_id(strategy_id)
    # print('strategy_id++++++++++++++++++++++++++++++=====',strategy_id)
    py_bwi_set_mode(mode)
    context.mode = mode
    context.strategy_id = strategy_id

    # 调用户文件的init
    context.inside_file_module = fmodule
    context.on_tick_fun = getattr(fmodule, 'on_tick', None)
    context.on_bar_fun = getattr(fmodule, 'on_bar', None)
    context.init_fun = getattr(fmodule, 'init', None)
    context.on_execution_report_fun = getattr(fmodule, 'on_execution_report', None)
    context.on_order_status_fun = getattr(fmodule, 'on_order_status', None)
    context.on_backtest_finished_fun = getattr(fmodule, 'on_backtest_finished', None)
    
    context.on_error_fun = getattr(fmodule, 'on_error', None)
    context.on_shutdown_fun = getattr(fmodule, 'on_shutdown', None)
    context.on_account_status_fun = getattr(fmodule, 'on_account_status', None)
    context.on_trade_data_connected_fun = getattr(fmodule, 'on_trade_data_connected', None)
    context.on_market_data_connected_fun = getattr(fmodule, 'on_market_data_connected', None)
    context.on_market_data_disconnected_fun = getattr(fmodule, 'on_market_data_disconnected', None)
    context.on_trade_data_disconnected_fun = getattr(fmodule, 'on_trade_data_disconnected', None)

    context.backtest_start_time = backtest_start_time
    context.backtest_end_time = backtest_end_time
    context.adjust_mode = backtest_adjust
    py_bwi_set_data_callback(callback_controller)  # 设置事件处理的回调函数

    splash_msgs = [
        '-' * 40,
        'python sdk version: {}'.format(__version__),
        '>>>'
        'cpp_sdk version: {}'.format(py_bwi_get_c_version()),
        '-' * 40,
    ]

    print(os.linesep.join(splash_msgs))

    if mode == MODE_BACKTEST:
        # beg="2019-06-10 17:20:00"
        # end="2019-07-10 17:30:00"
        # py_bwi_set_backtest_config(beg,end,1000000, 1, 0, 0, 1, 1)
        py_bwi_set_backtest_config(backtest_start_time,
                                   backtest_end_time,
                                   backtest_initial_cash,
                                   backtest_commission_ratio,
                                   backtest_slippage_ratio,
                                   backtest_adjust,
                                   backtest_check_cache
                                   )
        # try:
        return py_bwi_run()
        #     pass
        # except Exception as identifier:
        #     print('run over')
        #     print(identifier)
        #     pass
    else:
        ## set_cache(backtest_check_cache)
        pass

    if py_bwi_init():
        # print('------------eeeeeeeeeeeeeeeeeeeeeeeeee')
        return

    context._set_accounts()

    while running:
        # print('-'*100,'running')
        py_bwi_poll()


def get_parameters():
    # type: () ->List[Dict[Text, Any]]

    return None




def log(level, msg, source):
    # type: (Text, Text, Text) ->NoReturn
    logs = Logs()
    item = logs.data.add()  # type: Log
    item.owner_id = context.strategy_id
    item.source = source
    item.level = level
    item.msg = msg

    req = logs.SerializeToString()
    py_bwi_log(req)


def stop():
    """
    停止策略的运行,用exit(2)退出
    """
    # print("&"*79,"stop")
    global running
    running = False
    import platform
    if platform.system()=='Windows':
        # print(platform.system())
        sys.exit(2)

#not in use
def set_serv_addr(addr):
    # type: (Text) -> NoReturn
    """
    设置终端服务地址
    """
    # py_bwi_set_serv_addr(addr)



from typing import Text, Dict,List


# nipasbc
class SecurityType(object):
    SecurityType_Unknown = 0  # type: int
    SecurityType_Stock   = 1  # type: int
    SecurityType_Fund    = 2  # type: int
    SecurityType_Index   = 3  # type: int
    SecurityType_Future  = 4  # type: int
    SecurityType_Option  = 5  # type: int

# nipasbc
class Quote(object):
    # bid_p = ...  # type: float
    # bid_v = ...  # type: int
    # ask_p = ...  # type: float
    # ask_v = ...  # type: int
    def __init__(self, bid_p:float=0, bid_v:int=0, ask_p:float=0, ask_v:int=0): ...
    def __init__(self):
        self.bid_price=0#                            ///本档委买价
        self.bid_volume=0#                           ///本档委买量
        self.ask_price=0#                            ///本档委卖价
        self.ask_volume=0#
          
    def __init__(self,q):
        self.bid_price=q.    bid_price#                            ///本档委买价
        self.bid_volume=q.   bid_volume#                           ///本档委买量
        self.ask_price=q.    ask_price#                            ///本档委卖价
        self.ask_volume=q.   ask_volume#                           ///本档委卖量
        pass
    def FromFmt(self,q):
        self.bid_price=q.    bid_price#                            ///本档委买价
        self.bid_volume=q.   bid_volume#                           ///本档委买量
        self.ask_price=q.    ask_price#                            ///本档委卖价
        self.ask_volume=q.   ask_volume#                           ///本档委卖量
        pass

# nipasbc
class OrderBook(object):
    symbol = ... # type: Text
    quotes = ... # type: List[Quote]
    created_at = ... # type: int
    pass

    # def __init__(self, symbol:Text=0, quotes:List[Quote]=None, created_at:int=None): ...

# nipasbc
class OrderBooks(object):
    data = ... # type: List[OrderBook]
    pass

    # def __init__(self, data:List[OrderBook]): ...

# nipasbc
class Trade(object):
    # symbol = ... # type: Text
    # price = ... # type: float
    # last_volume = ... # type: int
    # last_amount = ... # type: float
    # cum_volume = ... # type: int
    # cum_amount = ... # type: float
    # trade_type = ... # type: int
    # created_at = ... # type: int

    def __init__(self,trade_):
        self.strategy_id=trade_.strategy_id#         //策略ID                                                    
        self.account_id=trade_.account_id#          //账号ID                            
        self.account_name=trade_.account_name#        //账户登录名                                                  
        self.cl_ord_id=trade_.cl_ord_id#          //委托客户端ID                                                 
        self.order_id=trade_.order_id#           //委托柜台ID                                                  
        self.exec_id=trade_.exec_id#           //委托回报ID                                                  
        self.symbol=trade_.symbol#          //symbol                                                    
        self.position_effect=trade_.position_effect#           //开平标志，取值参考enum PositionEffect                                     
        self.side=trade_.side#                 //买卖方向，取值参考enum OrderSide                                       
        self.ord_rej_reason=trade_.ord_rej_reason#            //委托拒绝原因，取值参考enum OrderRejectReason                                 
        self.ord_rej_reason_detail=trade_.ord_rej_reason_detail#   //委托拒绝原因描述                                               
        self.exec_type=trade_.exec_type#              //执行回报类型, 取值参考enum ExecType                                      
        self.price=trade_.price#                //委托成交价格                                                 
        self.volume=trade_.volume#                //委托成交量                                                  
        self.amount=trade_.amount#                //委托成交金额                                                 
        self.commission=trade_.commission#              //委托成交手续费                                                
        self.cost=trade_.cost#                 //委托成交成本金额 
        self.created_at=trade_.created_at#              //回报创建时间
    # def __init__(self, symbol:Text='', price:float=0, last_volume:int=0, last_amount:float=0,
    #              cum_volume:int=0, cum_amount:float=0, trade_type:int=0,
    #              created_at:int=None):...

# nipasbc
class Trades(object):
    data = ... # type: List[Trade]
    pass
    # def __init__(self, data:List[Trade]): ...

# nipasbc
class Tick(object):
    # symbol = ... # type: Text

    # open = ... # type: float
    # high = ... # type: float
    # low = ... # type: float
    # price = ... # type: float

    # quotes = ... # type: List[Quote]

    # cum_volume = ... # type: int
    # cum_amount = ... # type: float
    # cum_position = ... # type: int
    # last_amount = ... # type: float
    # last_volume = ... # type: int
    # trade_type = ... # type: int

    # created_at = ... # type: int
    def __init__(self):
        pass
    
    def __init__(self, symbol:Text='', open:float=0, high:float=0, low:float=0, price:float=0,
                 quotes:List[Quote]=None, cum_volume:int=0, cum_amount:float=0, cum_position:int=0,
                 last_amount:float=0, last_volume:int=0, trade_type:int=0, created_at:int=None): ...

    def __init__(self,t):
        self.symbol = t.symbol#
        self.created_at =  Timestamp(t.created_at)#               ///<utc时间，精确到毫秒--期货等
        self.price = t.price#                    ///<最新价
        self.open = t.open#                     ///<开盘价
        self.high = t.high#                     ///<最高价
        self.low = t.low#                      ///<最低价
        self.cum_volume = t.cum_volume#               ///<成交总量
        self.cum_amount = t.cum_amount#               ///<成交总金额/最新成交额,累计值
        self.cum_position = t.cum_position#             ///<合约持仓量(const 期),累计值
        self.last_amount = t.last_amount#              ///<瞬时成交额
        self.last_volume = t.last_volume#              ///<瞬时成交量
        self.trade_type = t.trade_type#               ///(const 保留)交易类型,对应多开,多平等类型
        
        self.quotes =[Quote(item) for item in t.quotes]    
        
        pass
    def FromFmt(self,t):
        self.symbol = t.symbol#
        self.created_at =  Timestamp(t.created_at)#               ///<utc时间，精确到毫秒--期货等
        self.price = t.price#                    ///<最新价
        self.open = t.open#                     ///<开盘价
        self.high = t.high#                     ///<最高价
        self.low = t.low#                      ///<最低价
        self.cum_volume = t.cum_volume#               ///<成交总量
        self.cum_amount = t.cum_amount#               ///<成交总金额/最新成交额,累计值
        self.cum_position = t.cum_position#             ///<合约持仓量(const 期),累计值
        self.last_amount = t.last_amount#              ///<瞬时成交额
        self.last_volume = t.last_volume#              ///<瞬时成交量
        self.trade_type = t.trade_type#               ///(const 保留)交易类型,对应多开,多平等类型
        
        self.quotes =[Quote(item) for item in t.quotes]  
        pass


    # def __init__(self, data:List[Tick]): ...

class Timestamp(object):
    # seconds=0
    # nanos=0
    def __init__(self):
        self.seconds = 0
        self.nanos = 0
        pass
    #double
    def __init__(self,t):
        if isinstance(t,float):
            self.seconds = int(t)
            self.nanos = int(t*1000)%1000*1000000000

        elif isinstance(t,int): #milisec
            self.seconds = int(t)/1000
            self.nanos = int(t)%1000*1000000000

        else:
            self.seconds = 0
            self.nanos = 0
        pass


# nipasbc
class Bar(object):
    # symbol = ... # type: Text
    # frequency = ... # type: Text

    # open = ... # type: float
    # high = ... # type: float
    # low = ... # type: float
    # close = ... # type: float
    # volume = ... # type: int
    # amount = ... # type: float
    # position = ... # type: int
    # pre_close = ... # type: float

    # bob = ... # type: int
    # eob = ... # type: int

    # def __init__(self,b):
    #     self.symbol = b.symbol#		        ///symbol
    #     self.bob = Timestamp(b.bob)#                     ///bar的开始时间
    #     self.eob = Timestamp(b.eob)#                        ///bar的结束时间
    #     self.open = b.open#                        ///<开盘价
    #     self.close = b.close#                        ///<收盘价
    #     self.high = b.high#                       ///<最高价
    #     self.low = b.low#                      ///<最低价
    #     self.volume = b.volume#                       ///<成交量
    #     self.amount = b.amount#                       ///<成交金额
    #     self.pre_close = b.pre_close#                    ///昨收盘价，只有日频数据赋值
    #     self.position = b.position#                     ///<持仓量
    #     self.frequency = b.frequency#	////bar频度
    #     pass
    def __init__(self):
        self.symbol = None         #///symbol
        self.bob = 0             #        ///bar的开始时间
        self.eob = 0              #       ///bar的结束时间
        self.open = 0 #                        ///<开盘价
        self.close = 0#                        ///<收盘价
        self.high = 0#                       ///<最高价
        self.low =0                #      ///<最低价
        self.volume = 0#                       ///<成交量
        self.amount = 0#                       ///<成交金额
        self.pre_close = 0#                    ///昨收盘价，只有日频数据赋值
        self.position = 0#                     ///<持仓量
        self.frequency = None#	////bar频度
        pass

    def __init__(self,b):
        self.symbol = b.symbol#		///symbol
        self.bob = Timestamp(b.bob)#                     ///bar的开始时间
        self.eob = Timestamp(b.eob)#                        ///bar的结束时间
        self.open = b.open#                        ///<开盘价
        self.close = b.close#                        ///<收盘价
        self.high = b.high#                       ///<最高价
        self.low = b.low#                      ///<最低价
        self.volume = b.volume#                       ///<成交量
        self.amount = b.amount#                       ///<成交金额
        self.pre_close = b.pre_close#                    ///昨收盘价，只有日频数据赋值
        self.position = b.position#                     ///<持仓量
        self.frequency = b.frequency#	////bar频度
        pass

    def __init__(self,b):
        self.symbol = b.symbol#		///symbol
        self.bob = Timestamp(b.bob)#                     ///bar的开始时间
        self.eob = Timestamp(b.eob)#                        ///bar的结束时间
        self.open = b.open#                        ///<开盘价
        self.close = b.close#                        ///<收盘价
        self.high = b.high#                       ///<最高价
        self.low = b.low#                      ///<最低价
        self.volume = b.volume#                       ///<成交量
        self.amount = b.amount#                       ///<成交金额
        self.pre_close = b.pre_close#                    ///昨收盘价，只有日频数据赋值
        self.position = b.position#                     ///<持仓量
        self.frequency = b.frequency#	////bar频度
        pass

    def FromFmt(self,b):
        self.symbol = b.symbol#		///symbol
        self.bob = Timestamp(b.bob)#                     ///bar的开始时间
        self.eob = Timestamp(b.eob)#                        ///bar的结束时间
        self.open = b.open#                        ///<开盘价
        self.close = b.close#                        ///<收盘价
        self.high = b.high#                       ///<最高价
        self.low = b.low#                      ///<最低价
        self.volume = b.volume#                       ///<成交量
        self.amount = b.amount#                       ///<成交金额
        self.pre_close = b.pre_close#                    ///昨收盘价，只有日频数据赋值
        self.position = b.position#                     ///<持仓量
        self.frequency = b.frequency#	////bar频度
        pass



# nipasbc
class InstrumentInfo(object):
    # symbol = ... # type: Text
    # sec_type = ... # type: int
    # exchange = ... # type: Text
    # sec_id = ... # type: Text
    # sec_name = ... # type: Text
    # sec_abbr = ... # type: Text
    # price_tick = ... # type: Text

    # listed_date = ... # type: int
    # delisted_date = ... # type: int
    pass


    # def __init__(self, symbol:Text='', sec_type:int=0, exchange:Text='', sec_id:Text='', sec_name:Text='',
    #              sec_abbr: Text='', price_tick: Text='', listed_date:int=None, delisted_date:int=None): ...

# nipasbc
class InstrumentInfos(object):
    data = ... # type: List[InstrumentInfo]

    # def __init__(self, data:List[InstrumentInfo]=None):...

# nipasbc
class Instrument(object):
    symbol = ... # type: Text
    sec_level = ... # type: int
    is_suspended = ... # type: int
    multiplier = ... # type: float
    margin_ratio = ... # type: float

    settle_price = ... # type: float
    position = ... # type: int
    pre_close = ... # type: float
    pre_settle = ... # type: float
    upper_limit = ... # type: float
    lower_limit = ... # type: float
    adj_factor = ... # type: float

    info = ... # type: InstrumentInfo

    created_at = ... # type: int

    # def __init__(self, symbol:Text='', sec_level:int=0, is_suspended:int=0, multiplier:float=0, margin_ratio:float=0,
    #              settle_price:float=0, position:int=0, pre_close:float=0, pre_settle:float=0, upper_limit:float=0, lower_limit:float=0,
    #              adj_factor:float=0, info: InstrumentInfo=None, created_at:int=None): ...

# nipasbc
class Instruments(object):
    data = ... # type: List[Instrument]

    # def __init__(self, data:List[Instrument]=None): ...

# nipasbc
class Dividend(object):
    symbol = ... # type: Text
    cash_div = ... # type: float
    share_div_ratio = ... # type: float
    share_trans_ratio = ... # type: float
    allotment_ratio = ... # type: float
    allotment_price = ... # type: float

    created_at = ... # type: int

    def __init__(self, symbol:Text='', cash_div:float=0, share_div_ratio:float=0, share_trans_ratio:float=0,
                 allotment_ratio:float=0, allotment_price:float=0): ...

# nipasbc
class Dividends(object):
    data = ... # type: List[Dividend]

    # def __init__(self, data:List[Dividend]=None): ...

# # nipasbc
# class ContinuousContract(object):
#     symbol = ... # type: Text
#     created_at = ... # type: int

#     # def __init__(self, symbol:Text='', created_at:int=None): ...

# # nipasbc
# class ContinuousContracts(object):
#     data = ... # type: List[ContinuousContract]

#     # def __init__(self, data=List[ContinuousContract]): ...

# nipasbc
class Constituent(object):
    constituents = ... # type: Dict[Text, float]
    created_at = ... # type: int

    # def __init__(self, constituents:Dict[Text, float], created_at:int): ...

# nipasbc
class Constituents(object):
    data = ... # type: List[Constituent]

    # def __init__(self, data:List[Constituent]=None): ...



# coding=utf-8
from __future__ import unicode_literals, print_function, absolute_import

# import grpc
import pandas as pd

from bw.constant import HISTORY_ADDR
from bw.csdk.cpp_sdk import py_bwi_get_serv_addr
from bw.enum import ADJUST_NONE,ADJUST_PREV,ADJUST_POST
from bw.model.storage import context

from bw.yy.data import *

from bw.utils import standard_fields,timestamp2bj_datetime,to_datetime_bjtzinfo

from bw.csdk.cpp_sdk import py_bwi_history_ticks, py_bwi_history_bars_n,py_bwi_get_trading_dates,py_bwi_history_bars,py_bwi_history_bars2,py_bwi_history_ticks_n

import datetime 

from bw.model.fundamental import FundamentalApi
fundamentalapi = FundamentalApi()

MAX_MESSAGE_LENGTH = 1024 * 1024 * 128

class HistoryApi(object):
    def __init__(self):
        self.addr = None

    def _init_addr(self):
        addr = py_bwi_get_serv_addr(HISTORY_ADDR)
        if not addr:
            raise EnvironmentError("获取不到数据服务地址")

        if not self.addr:
            self.addr = addr

    def reset_addr(self):
        self.addr = None

    def get_history_bars(self, symbol, frequency, start_time, end_time,
                         fields=None,
                         adjust=ADJUST_PREV, adjust_end_time='', df=False):
                        
        
        cdatas = py_bwi_history_bars(symbol,frequency,start_time, end_time ,adjust)

        if cdatas is None:
            return None
        datas = []
        for item in cdatas:
            bar = Bar(item)
            bar.FromFmt(item)
            barnew = {
                'symbol': bar.symbol,
                'eob': timestamp2bj_datetime(bar.eob),
                'bob': timestamp2bj_datetime(bar.bob),
                'open': bar.open,
                'close': bar.close,
                'high': bar.high,
                'low': bar.low,
                'volume': bar.volume,
                'amount': bar.amount,
                'pre_close': bar.pre_close,
                'position': bar.position,
                'frequency': bar.frequency,
                # 'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            datas.append(barnew)
            pass
        
        # print("cdata------->",datas)

        fields_str, fields_list = standard_fields(fields, letter_upper=False)
        if not fields_list:
            datas = datas if not df else pd.DataFrame(datas)
            return datas
        else:
            result = []
            for data in datas:
                info = {}
                for field in fields_list:
                    info[field] = data.get(field)
                result.append(info)

            result = result if not df else pd.DataFrame(result)
            return result
        return cdatas


    def get_history_bars2(self, symbol, frequency, start_time, end_time,
                         fields=None,
                         adjust=ADJUST_PREV, adjust_end_time='', df=False):
                        
        
        cdatas = py_bwi_history_bars2(symbol,frequency,start_time, end_time ,adjust)

        if cdatas is None:
            return None
        datas = []
        for item in cdatas:
            bar = Bar(item)
            bar.FromFmt(item)
            barnew = {
                'symbol': bar.symbol,
                'eob': timestamp2bj_datetime(bar.eob),
                'bob': timestamp2bj_datetime(bar.bob),
                'open': bar.open,
                'close': bar.close,
                'high': bar.high,
                'low': bar.low,
                'volume': bar.volume,
                'amount': bar.amount,
                'pre_close': bar.pre_close,
                'position': bar.position,
                'frequency': bar.frequency,
                # 'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            datas.append(barnew)
            pass
        
        # print("cdata------->",datas)

        fields_str, fields_list = standard_fields(fields, letter_upper=False)
        if not fields_list:
            datas = datas if not df else pd.DataFrame(datas)
            return datas
        else:
            result = []
            for data in datas:
                info = {}
                for field in fields_list:
                    info[field] = data.get(field)
                result.append(info)

            result = result if not df else pd.DataFrame(result)
            return result
        return cdatas


    def get_history_ticks(self, symbols, start_time, end_time, fields=None,
                          skip_suspended=True, fill_missing=None,
                          adjust=ADJUST_NONE, adjust_end_time='', df=False):
        
   
        cdata =py_bwi_history_ticks(symbols,start_time,end_time, adjust, adjust_end_time, skip_suspended, fill_missing)

        datas = []
        for item in cdatas:
            tick = Tick()
            tick.FromFmt(item)
            if len(tick.quotes) < 5:
                for _ in range(len(tick.quotes), 5):
                    zero_val = {'bid_p': 0, 'bid_v': 0, 'ask_p': 0, 'ask_v': 0}
                    tick.quotes.append(zero_val)

            ticknew = {
                'quotes': quotes,
                'symbol': tick.symbol,
                'created_at': timestamp2bj_datetime(tick.created_at),
                'price': tick.price,
                'open': tick.open,
                'high': tick.high,
                'low': tick.low,
                'cum_volume': tick.cum_volume,
                'cum_amount': tick.cum_amount,
                'cum_position': tick.cum_position,
                'last_amount': tick.last_amount,
                'last_volume': tick.last_volume,
                'trade_type': tick.trade_type,
                'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            datas.append(ticknew)
            pass
        
        fields_str, fields_list = standard_fields(fields, letter_upper=False)
        if not fields_list:
            datas = datas if not df else pd.DataFrame(datas)
            return datas
        else:
            result = []
            for data in datas:
                info = {}
                for field in fields_list:
                    info[field] = data.get(field)
                result.append(info)

            result = result if not df else pd.DataFrame(result)
            return result
        return cdatas

    # KLINE N py_bwi_history_bars_n
    def get_history_n_bars3(self, symbol, frequency, count, end_time= datetime.datetime.now(),
                           fields=None,
                           adjust=ADJUST_PREV,  fill_missing=None,df=False):
 
        start_back_with = 0
        nowtime = to_datetime_bjtzinfo (datetime.datetime.now() ) # to_datetime_bjtzinfo(end_time)
        endtime = to_datetime_bjtzinfo(end_time)
        
        codelist = symbol.split('.')

        # print('-'*20,codelist,'==',endtime.strftime("%Y-%m-%d"),nowtime.strftime("%Y-%m-%d"))
        
        trading_day_ls = fundamentalapi.get_trading_dates(codelist[0],endtime.strftime("%Y-%m-%d"),nowtime.strftime("%Y-%m-%d"))
        
        start_back_with =  len(trading_day_ls) -1 if len(trading_day_ls)>0 else 0

        # print(start_back_with,'-'*60,nowtime,end_time)

        cdatas = py_bwi_history_bars_n(symbol,frequency,count,start_back_with,adjust)
        
        datas = []
        for item in cdatas:
            bar = Bar(item)
            bar.FromFmt(item)
            # print(bar.bob)
            barnew = {
                'symbol': bar.symbol,
                'eob': timestamp2bj_datetime(bar.eob),
                'bob': timestamp2bj_datetime(bar.bob),
                'open': bar.open,
                'close': bar.close,
                'high': bar.high,
                'low': bar.low,
                'volume': bar.volume,
                'amount': bar.amount,
                'pre_close': bar.pre_close,
                'position': bar.position,
                'frequency': bar.frequency,
                # 'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            datas.append(barnew)
            pass
        
        # print('over construct!!')
        # print(fields)

        
        fields_str, fields_list = standard_fields(fields, letter_upper=False)
        
        # print('over standard_fields!!', fields_str,fields_list)
        if not fields_list:
            datas = datas if not df else pd.DataFrame(datas)
            return datas
        else:
            result = []
            for data in datas:
                info = {}
                for field in fields_list:
                    info[field] = data.get(field)
                result.append(info)

            result = result if not df else pd.DataFrame(result)
            return result
        return cdatas

    
    
    #KLINE DATE  by N py_bwi_history_bars2
    def get_history_n_bars2(self, symbol, frequency, count, end_time= datetime.datetime.now(),
                           fields=None,
                           adjust=ADJUST_PREV,  fill_missing=None,df=False):
                           
        start_back_with = 0
        nowtime = to_datetime_bjtzinfo (datetime.datetime.now() ) # to_datetime_bjtzinfo(end_time)
        endtime = to_datetime_bjtzinfo(end_time)
        
        codelist = symbol.split('.')

        year_plus = datetime.timedelta(days=int(365*(count+365)/245))
        # print('year---plus ',year_plus,count)
        begtime = endtime - year_plus
        # print('------begtime----endtime',begtime,' ',endtime)
        # print('-'*20,codelist,'==',endtime.strftime("%Y-%m-%d"),nowtime.strftime("%Y-%m-%d"))
        
        trading_day_ls = fundamentalapi.get_trading_dates(codelist[0],begtime.strftime("%Y-%m-%d"),endtime.strftime("%Y-%m-%d"))
        lentradingdays = len(trading_day_ls)
        # print('len trading_day_ls:',lentradingdays)
        if trading_day_ls is None:
            return []
        time_start = begtime.strftime("%Y-%m-%d")
        if lentradingdays > int(count) :
            time_start = trading_day_ls[-count]
            # print('----seg time: ',time_start)
            # trading_day_ls_new = fundamentalapi.get_trading_dates(codelist[0],time_start,endtime.strftime("%Y-%m-%d"))
            # print(len(trading_day_ls_new))
        else:
            print('get_history_n_bars2 : count error occured')

        # print(start_back_with,'-'*60,nowtime,end_time)

        cdatas = py_bwi_history_bars2(symbol,frequency,time_start, endtime.strftime("%Y-%m-%d") ,adjust)
        
        datas = []
        for item in cdatas:
            bar = Bar(item)
            bar.FromFmt(item)
            # print(bar.bob)
            barnew = {
                'symbol': bar.symbol,
                'eob': timestamp2bj_datetime(bar.eob),
                'bob': timestamp2bj_datetime(bar.bob),
                'open': bar.open,
                'close': bar.close,
                'high': bar.high,
                'low': bar.low,
                'volume': bar.volume,
                'amount': bar.amount,
                'pre_close': bar.pre_close,
                'position': bar.position,
                'frequency': bar.frequency,
                # 'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            datas.append(barnew)
            pass
        
        # print('over construct!!')
        # print(fields)

        
        fields_str, fields_list = standard_fields(fields, letter_upper=False)
        
        # print('over standard_fields!!', fields_str,fields_list)
        if not fields_list:
            datas = datas if not df else pd.DataFrame(datas)
            return datas
        else:
            result = []
            for data in datas:
                info = {}
                for field in fields_list:
                    info[field] = data.get(field)
                result.append(info)

            result = result if not df else pd.DataFrame(result)
            return result
        return cdatas

    
    #KLINE BY DATE py_bwi_history_bars
    def get_history_n_bars(self, symbol, frequency, count, end_time= datetime.datetime.now(),
                           fields=None,
                           adjust=ADJUST_PREV,  fill_missing=None,df=False):
 
        start_back_with = 0
        nowtime = to_datetime_bjtzinfo (datetime.datetime.now() ) # to_datetime_bjtzinfo(end_time)
        endtime = to_datetime_bjtzinfo(end_time)
        
        codelist = symbol.split('.')

        year_plus = datetime.timedelta(days=int(365*(count+365)/245))
        # print('year---plus ',year_plus,count)
        begtime = endtime - year_plus
        # print('------begtime----endtime',begtime,' ',endtime)
        # print('-'*20,codelist,'==',begtime,endtime.strftime("%Y-%m-%d"),nowtime.strftime("%Y-%m-%d"))
        
        trading_day_ls = fundamentalapi.get_trading_dates(codelist[0],begtime.strftime("%Y-%m-%d"),endtime.strftime("%Y-%m-%d"))
        lentradingdays = len(trading_day_ls)
        # print('len trading_day_ls:', lentradingdays)
        if trading_day_ls is None:
            return []
        time_start = begtime.strftime("%Y-%m-%d")
        if lentradingdays > int(count) :
            time_start = trading_day_ls[-count]
            # print('----seg time: ',time_start)
            trading_day_ls_new = fundamentalapi.get_trading_dates(codelist[0],time_start,endtime.strftime("%Y-%m-%d"))
            # print(len(trading_day_ls_new))
        else:
            print('get_history_n_bars count error occured')

        # print(start_back_with,'-'*60,nowtime,end_time)
        # print(symbol,frequency,time_start, endtime.strftime("%Y-%m-%d") ,adjust)
        cdatas = py_bwi_history_bars(symbol,frequency,time_start, endtime.strftime("%Y-%m-%d") ,adjust)
        # print(cdatas)
        
        datas = []
        for item in cdatas:
            bar = Bar(item)
            bar.FromFmt(item)
            # print(bar.bob)
            barnew = {
                'symbol': bar.symbol,
                'eob': timestamp2bj_datetime(bar.eob),
                'bob': timestamp2bj_datetime(bar.bob),
                'open': bar.open,
                'close': bar.close,
                'high': bar.high,
                'low': bar.low,
                'volume': bar.volume,
                'amount': bar.amount,
                'pre_close': bar.pre_close,
                'position': bar.position,
                'frequency': bar.frequency,
                # 'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            datas.append(barnew)
            pass
        
        # print('over construct!!')
        # print(fields)

        
        fields_str, fields_list = standard_fields(fields, letter_upper=False)
        
        # print('over standard_fields!!', fields_str,fields_list)
        if not fields_list:
            datas = datas if not df else pd.DataFrame(datas)
            return datas
        else:
            result = []
            for data in datas:
                info = {}
                for field in fields_list:
                    info[field] = data.get(field)
                result.append(info)

            result = result if not df else pd.DataFrame(result)
            return result
        return cdatas

    

    def get_history_n_ticks(self, symbol, count, end_time=None, fields=None,
                            skip_suspended=None,
                            fill_missing=None, adjust=ADJUST_NONE,
                            adjust_end_time='', df=False):
        
        cdata =py_bwi_history_ticks_n(symbols, n, end_time,adjust, adjust_end_time,  skip_suspended, fill_missing)
        datas = []
        for item in cdatas:
            tick = Tick()
            tick.FromFmt(item)
            if len(tick.quotes) < 5:
                for _ in range(len(tick.quotes), 5):
                    zero_val = {'bid_p': 0, 'bid_v': 0, 'ask_p': 0, 'ask_v': 0}
                    tick.quotes.append(zero_val)

            ticknew = {
                'quotes': quotes,
                'symbol': tick.symbol,
                'created_at': timestamp2bj_datetime(tick.created_at),
                'price': tick.price,
                'open': tick.open,
                'high': tick.high,
                'low': tick.low,
                'cum_volume': tick.cum_volume,
                'cum_amount': tick.cum_amount,
                'cum_position': tick.cum_position,
                'last_amount': tick.last_amount,
                'last_volume': tick.last_volume,
                'trade_type': tick.trade_type,
                'receive_local_time': time.time(),  # 收到时的本地时间秒数
            }
            datas.append(ticknew)
            pass
        
        fields_str, fields_list = standard_fields(fields, letter_upper=False)
        if not fields_list:
            datas = datas if not df else pd.DataFrame(datas)
            return datas
        else:
            result = []
            for data in datas:
                info = {}
                for field in fields_list:
                    info[field] = data.get(field)
                result.append(info)

            result = result if not df else pd.DataFrame(result)
            return result
        return cdatas

        


# coding=utf-8
from __future__ import unicode_literals, print_function, absolute_import

from datetime import date as Date, datetime as Datetime

# import grpc
import six
# from google.protobuf.timestamp_yy2 import int
from six import string_types
from typing import List, Dict, Text, Any, Union

from bw import utils
from bw.constant import FUNDAMENTAL_ADDR
from bw.csdk.cpp_sdk import py_bwi_get_serv_addr
from bw.model.storage import context


# from bw.yy.data_yy2 import Instrument, InstrumentInfo, ContinuousContract, Dividend

from bw.csdk.cpp_sdk import py_bwi_current,py_bwi_log, py_bwi_strerror, \
    py_bwi_run, py_bwi_set_serv_addr,\
     py_bwi_init, py_bwi_poll, py_bwi_get_c_version, py_bwi_set_apitoken,\
    py_bwi_get_sector,py_bwi_get_trading_dates,py_bwi_get_next_trading_date,\
        py_bwi_get_previous_trading_date,py_bwi_get_fundamentals,\
            py_bwi_get_industry,py_bwi_get_constituents


from bw.utils import str_lowerstrip, load_to_list

BwDate = Union[Text, Datetime, Date]  # 自定义bw里可表示时间的类型
TextNone = Union[Text, None]  # 可表示str或者None类型

MAX_MESSAGE_LENGTH = 1024 * 1024 * 128



def get_sec_type_str(sec_types):
    """把int类型的sectype转为字符串的sectype, 不能转换则返回None"""
    d = {
        1: 'stock',
        2: 'bond',
        3: 'fund',
        4: 'warrant',
        5: 'index',
        6:'futures',# 股指
        7:'stock_kc',
        20:'plate',
        21:'concept',

        '1': 'stock',
        '2': 'bond',
        '3': 'fund',
        '4': 'warrant',
        '5': 'index',
        '6':'futures',# 股指
        '7':'stock_kc',
        '20':'plate',
        '21':'concept',

        'stock':'stock',
        'bond':'bond',
        'fund':'fund',
        'warrant':'warrant',
        'index':'index',
        'futures':'futures',
        'stock_kc':'stock_kc',
        'plate':'plate',
        'concept':'concept',
    }
    result = []
    for sec_type in sec_types:
        if isinstance(sec_type, six.string_types):
            sec_type = sec_type.strip().lower()
        if sec_type in d:
            result.append(d.get(sec_type))

    return result


class FundamentalApi(object):
    def __init__(self):
        self.addr = None

    def _init_addr(self):
        # new_addr = py_bwi_get_serv_addr(FUNDAMENTAL_ADDR)
        # if not new_addr:
        #     raise EnvironmentError("获取不到基本面服务地址")

        # if not self.addr:
        #     self.addr = new_addr
        pass

    def reset_addr(self):
        # self.addr = None
        pass
    #support seperate industry indicate
    def get_fundamentals(self,symbols,  collection ,field,start_date, 
    end_date,limit=1000,  table="SHSZADB"):
        """
        查询财务以及相关数据，详细见文档说明
        """
        

        if isinstance(symbols, string_types):
            symbols = [s.strip() for s in symbols.split(',') if s.strip()]
            symbols = ','.join(symbols)
        
        result = py_bwi_get_fundamentals(symbols,  collection ,field,start_date, end_date,limit,  table)
        # print('-'*100,result)
        return result

    def get_instruments(self, symbols=None, exchanges=None, sec_types=None, names=None, skip_suspended=True,
                        skip_st=True, fields=None):
        """
        查询最新交易标的信息,有基本数据及最新日频数据
        """
        # self._init_addr()
        
        instrument_fields = {
            'symbol', 'sec_level', 'is_suspended', 'multiplier', 'margin_ratio',
            'settle_price',
            'position', 'pre_close', 'upper_limit', 'lower_limit', 'adj_factor',
            'created_at', 'trade_date'
        }

        info_fields = {
            'sec_type', 'exchange', 'sec_id', 'sec_name', 'sec_abbr',
            'price_tick','listed_date', 'delisted_date'
        }

        all_fields = instrument_fields.union(info_fields)

        if isinstance(symbols, string_types):
            symbols = [s for s in map(str_lowerstrip, symbols.split(',')) if s]
        if not symbols:
            symbols = []

        if isinstance(exchanges, string_types):
            exchanges = [utils.to_exchange(s) for s in exchanges.split(',') if utils.to_exchange(s)]
        if not exchanges:
            exchanges = []

        if isinstance(sec_types, six.string_types):
            sec_types = [it.strip() for it in sec_types.split(',') if it.strip()]

        if isinstance(sec_types, int):
            sec_types = [sec_types]

        if isinstance(sec_types, list):
            sec_types = get_sec_type_str(sec_types)

        if not sec_types:
            sec_types = []

        if isinstance(names, string_types):
            names = [s for s in names.split(',') if s]
        if not names:
            names = []

        if not fields:
            filter_fields = all_fields
        elif isinstance(fields, string_types):
            filter_fields = {f for f in map(str_lowerstrip, fields.split(','))
                             if f in all_fields}
        else:
            filter_fields = {f for f in map(str_lowerstrip, fields) if
                             f in all_fields}

        if 'trade_date' in filter_fields:
            filter_fields.add('created_at')

        if not filter_fields:
            return []

        
        instrument_copy_field = filter_fields & instrument_fields
        info_copy_field = filter_fields & info_fields

        for ins in resp.data:  # type: Instrument
            row = dict()

            created_at_val = row.get('created_at', None)
            if isinstance(created_at_val, Datetime):
                row['trade_date'] = utils.utc_datetime2beijing_datetime(created_at_val)
                row.pop('created_at')

            listed_date_val = row.get('listed_date', None)
            if isinstance(listed_date_val, Datetime):
                row['listed_date'] = utils.utc_datetime2beijing_datetime(listed_date_val)

            delisted_date_val = row.get('delisted_date', None)
            if isinstance(delisted_date_val, Datetime):
                row['delisted_date'] = utils.utc_datetime2beijing_datetime(delisted_date_val)

            result.append(row)
        return result

    def get_history_instruments(self, symbols, fields=None, start_date=None, end_date=None):
        """
        返回指定的symbols的标的日指标数据
        """
        # self._init_addr()
        symbols = load_to_list(symbols)
        start_date = utils.to_datestr(start_date)
        end_date = utils.to_datestr(end_date)

        if not start_date:
            start_date = ''
        if not end_date:
            end_date = ''

        result=[]
        for info in result:
            created_at_val = info.get('created_at', None)
            if isinstance(created_at_val, Datetime):
                info['trade_date'] = utils.utc_datetime2beijing_datetime(created_at_val)
                info.pop('created_at')
        return result

    def get_instrumentinfos(self, symbols=None, exchanges=None, sec_types=None, names=None, fields=None):
        """
        查询交易标的基本信息
        如果没有数据的话,返回空列表. 有的话, 返回list[dict]这样的列表. 其中 listed_date, delisted_date 为 datetime 类型
        @:param fields: 可以是 'symbol, sec_type' 这样的字符串, 也可以是 ['symbol', 'sec_type'] 这样的字符list
        """
        # self._init_addr()

        if isinstance(symbols, string_types):
            symbols = [s for s in symbols.split(',') if s]
        if not symbols:
            symbols = []

        all_fields = {
            'symbol', 'sec_type', 'exchange', 'sec_id', 'sec_name', 'sec_abbr', 'price_tick', 'listed_date',
            'delisted_date'
        }

        if not fields:
            filter_fields = all_fields
        elif isinstance(fields, string_types):
            filter_fields = {f for f in map(str_lowerstrip, fields.split(',')) if f in all_fields}
        else:
            filter_fields = [f for f in map(str_lowerstrip, fields) if f in all_fields]

        if not filter_fields:
            return []

        if isinstance(exchanges, string_types):
            exchanges = [utils.to_exchange(s) for s in exchanges.split(',') if utils.to_exchange(s)]
        if not exchanges:
            exchanges = []

        if isinstance(sec_types, six.string_types):
            sec_types = [it.strip() for it in sec_types.split(',') if it.strip()]

        if isinstance(sec_types, int):
            sec_types = [sec_types]

        if isinstance(sec_types, list):
            sec_types = get_sec_type_str(sec_types)

        if not sec_types:
            sec_types = []

        if isinstance(names, string_types):
            names = [s for s in names.split(',') if s]
        if not names:
            names = []

        result = []
        for ins in resp.data:  # type: InstrumentInfo
            row = dict()
            
            listed_date_val = row.get('listed_date', None)
            if isinstance(listed_date_val, Datetime):
                row['listed_date'] = utils.utc_datetime2beijing_datetime(listed_date_val)

            delisted_date_val = row.get('delisted_date', None)
            if isinstance(delisted_date_val, Datetime):
                row['delisted_date'] = utils.utc_datetime2beijing_datetime(delisted_date_val)

            result.append(row)

        return result

    # def get_history_constituents(self, index, start_date=None, end_date=None):
    #     # type: (TextNone, BwDate, BwDate) -> List[Dict[Text, Any]]
    #     """
    #     查询指数历史成分股
    #     返回的list每项是个字典,包含的key值有:
    #     trade_date: 交易日期(datetime类型)
    #     constituents: 一个字典. 每个股票的sybol做为key值, weight做为value值
    #     """
    #     self._init_addr()

    #     start_date = utils.to_datestr(start_date)
    #     end_date = utils.to_datestr(end_date)

    #     if not start_date:
    #         start_date = Date.today()
    #     else:
    #         start_date = Datetime.strptime(start_date, '%Y-%m-%d').date()

    #     if not end_date:
    #         end_date = Date.today()
    #     else:
    #         end_date = Datetime.strptime(end_date, '%Y-%m-%d').date()
    #     return [
    #         {'trade_date': utils.utc_datetime2beijing_datetime(item.created_at.ToDatetime()),
    #          'constituents': dict(item.constituents)}
    #         for item in resp.data
    #     ]

    def get_constituents(self, index, fields=None, df=False):
        """
        查询指数最新成分股. 指定 fields = 'symbol, name'
        返回的list每项是个字典,包含的key值有:
        symbol 股票symbol
        
        如果不指定 fields, 则返回的list每项是symbol字符串
        """
        # self._init_addr()

        all_fields = ['symbol', 'name']
        if not fields:
            filter_fields = {'symbol'}
        elif isinstance(fields, string_types):
            filter_fields = {f for f in map(str_lowerstrip, fields.split(',')) if f in all_fields}
        else:
            filter_fields = {f for f in map(str_lowerstrip, fields) if f in all_fields}

        resp = py_bwi_get_constituents(index)
        if len(resp) > 0:
            filter_fields = list(filter_fields)
            if len(filter_fields) == 1 and filter_fields[0] == 'symbol':
                if not df:
                    return [k['symbol'] for k in resp]
                else:
                    return resp
            else:
                return resp
        else:
            return []

    def get_plates(self, ty):
        """
        查询板块股票列表  0 hy gn dy  1 hy 2 gn 3 dy 
        返回的list每项是个字典,包含的key值有:
        symbol 股票symbol
        """
        rsp = py_bwi_get_sector(ty)
        return rsp

    def get_industry(self, code):
        """
        查询行业股票列表
        """
        if not code:
            return []
        rsp = py_bwi_get_industry(code)
      
        return rsp

    def get_concept(self, code):
        """
        查询概念股票列表
        """
        if not code:
            return []
        rsp = py_bwi_get_industry(code)
      
        return rsp

    def get_trading_dates(self, exchange, start_date, end_date):
        # type: (Text, BwDate, BwDate) -> List[Text]
        """
        查询交易日列表
        如果指定的市场不存在, 返回空列表. 有值的话,返回 yyyy-mm-dd 格式的列表
        """

        exchange = utils.to_exchange(exchange)
        sdate = utils.to_datestr(start_date)
        edate = utils.to_datestr(end_date)
        if not exchange:
            return []
        if not sdate:
            return []
        if not end_date:
            edate = Datetime.now().strftime('%Y-%m-%d')
        # print(exchange,start_date,end_date)
        resp =  py_bwi_get_trading_dates(exchange,sdate,edate)
        # print('-'*20,resp)
        return resp

    def get_previous_trading_date(self, exchange, date):
        # type: (Text, BwDate) -> TextNone
        """
        返回指定日期的上一个交易日
        @:param exchange: 交易市场
        @:param date: 指定日期, 可以是datetime.date 或者 datetime.datetime 类型. 或者是 yyyy-mm-dd 或 yyyymmdd 格式的字符串
        @:return 返回下一交易日, 为 yyyy-mm-dd 格式的字符串, 如果不存在则返回None
        """
        exchange = utils.to_exchange(exchange)
        date_str = utils.to_datestr(date)
        if not exchange or not date_str:
            return None
        resp =  py_bwi_get_previous_trading_date(exchange,date)
        return resp

    def get_next_trading_date(self, exchange, date):
        # type: (Text, BwDate) -> TextNone
        """
        返回指定日期的下一个交易日
        @:param exchange: 交易市场
        @:param date: 指定日期, 可以是datetime.date 或者 datetime.datetime 类型. 或者是 yyyy-mm-dd 或 yyyymmdd 格式的字符串
        @:return 返回下一交易日, 为 yyyy-mm-dd 格式的字符串, 如果不存在则返回None
        """

        exchange = utils.to_exchange(exchange)
        date_str = utils.to_datestr(date)
        if not date_str or not exchange:
            return None
        resp =  py_bwi_get_next_trading_date(exchange,date)
        return resp

    #not support
    def get_dividend(self, symbol, start_date, end_date=None):
        # type: (Text, BwDate, BwDate) -> List[Dict[Text, Any]]
        """
        查询分红送配
        """

        if not symbol or not start_date:
            return []
        sym_tmp = symbol.split('.')  # List[Text]
        sym_tmp[0] = sym_tmp[0].upper()
        symbol = '.'.join(sym_tmp)

        if not end_date:
            end_date = Datetime.now().strftime('%Y-%m-%d')
        start_date = utils.to_datestr(start_date)
        end_date = utils.to_datestr(end_date)

        resp =[]
        result = []
        fields = ['symbol', 'cash_div', 'share_div_ratio', 'share_trans_ratio', 'allotment_ratio', 'allotment_price',
                  'created_at']
        for divi in resp.data:  # type: Dividend
            row = dict()
            created_at_val = row.get('created_at', None)
            if isinstance(created_at_val, Datetime):
                row['created_at'] = utils.utc_datetime2beijing_datetime(created_at_val)
            result.append(row)
        return result
    #not support
    def get_continuous_contracts(self, csymbol, start_date=None, end_date=None):
        # type: (Text, BwDate, BwDate) -> List[Dict[Text, Any]]

        start_date = utils.to_datestr(start_date)
        end_date = utils.to_datestr(end_date)

        resp =[]

        result = []
        for cc in resp.data:  # type: ContinuousContract
            row = {'symbol': cc.symbol, 'trade_date': utils.utc_datetime2beijing_datetime(cc.created_at.ToDatetime())}
            result.append(row)
        return result
    #not support
    def get_fundamentals_n(self, table, symbols, end_date, fields=None, filter=None, order_by=None, count=1):
        """
        查询基本面财务数据,每个股票在end_date的前n条
        """

        end_date = utils.to_datestr(end_date)

        if isinstance(symbols, string_types):
            symbols = [s.strip() for s in symbols.split(',') if s.strip()]

        # resp = GetFundamentalsNReq(table=table, end_date=end_date, fields=fields,
        #                           symbols=','.join(symbols), filter=filter,
        #                           order_by=order_by, count=count)

        resp=[] 
        result = []
        for item in resp.data:  # type: GetFundamentalsRsp.Fundamental
            r = {
                'symbol': item.symbol,
                'pub_date': utils.utc_datetime2beijing_datetime(item.pub_date.ToDatetime()),
                'end_date': utils.utc_datetime2beijing_datetime(item.end_date.ToDatetime()),
            }
            r.update(item.fields)
            result.append(r)

        return result
